"""Cluster analysis module."""

from collections import defaultdict
from pathlib import Path

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec
from matplotlib.image import NonUniformImage
from numpy import typing as npt

from irradiapy import dtypes, utils
from irradiapy.io.lammpsreader import LAMMPSReader
from irradiapy.io.lammpswriter import LAMMPSWriter

# region Clustering


def clusterize(
    defects: dtypes.Atom, cutoff: float
) -> tuple[dtypes.Acluster, dtypes.Ocluster]:
    """Identify atom clusters.

    Note
    ----
    Atom clusters are the individual atoms with their cluster number, while object clusters are a
    single point representing the average position of the atoms in the cluster and the type of the
    cluster (the cluster type is taken from the first atom in the cluster).

    Parameters
    ----------
    atoms : dtypes.Atom
        Array of atoms with fields "type", "x", "y", "z".
    cutoff : float
        Cutoff distance for clustering.

    Returns
    -------
    tuple[dtypes.Acluster, dtypes.Ocluster]
        Atomic and object clusters.
    """
    cutoff2 = cutoff**2
    natoms = defects.size
    aclusters = np.empty(natoms, dtype=dtypes.acluster)
    aclusters["x"] = defects["x"]
    aclusters["y"] = defects["y"]
    aclusters["z"] = defects["z"]
    aclusters["type"] = defects["type"]

    # Each atom is its own cluster initially
    aclusters["cluster"] = np.arange(1, natoms + 1)

    # Cluster identification
    for i in range(natoms):
        curr_cluster = aclusters[i]["cluster"]
        dists = (
            np.square(aclusters[i]["x"] - aclusters["x"][i + 1 :])
            + np.square(aclusters[i]["y"] - aclusters["y"][i + 1 :])
            + np.square(aclusters[i]["z"] - aclusters["z"][i + 1 :])
        )
        neighbours = aclusters["cluster"][np.where(dists <= cutoff2)[0] + i + 1]
        if neighbours.size:
            for neighbour in neighbours:
                if neighbour != curr_cluster:
                    aclusters["cluster"][
                        aclusters["cluster"] == neighbour
                    ] = curr_cluster

    # Rearrange cluster numbers to be consecutive starting from 1
    nclusters = np.unique(aclusters["cluster"])
    for i in range(nclusters.size):
        aclusters["cluster"][aclusters["cluster"] == nclusters[i]] = i + 1

    # Calculate object clusters
    oclusters = atom_to_object(aclusters)

    return aclusters, oclusters


def clusterize_file(
    path_defects: Path,
    cutoff_sia: float,
    cutoff_vac: float,
    path_aclusters: Path,
    path_oclusters: Path,
) -> None:
    """Finds defect clusters in the given file.

    Parameters
    ----------
    path_defects : Path
        Path of the file where defects are.
    cutoff_sia : float
        Cutoff distance for interstitials clustering.
    cutoff_vac : float
        Cutoff distance for vacancies clustering.
    path_aclusters : Path
        Where atomic clusters will be stored.
    path_oclusters : Path
        Where object clusters will be stored.
    """
    reader = LAMMPSReader(path_defects)
    nsim = 0
    with (
        LAMMPSWriter(path_aclusters) as awriter,
        LAMMPSWriter(path_oclusters) as owriter,
    ):
        for data_defects in reader:
            nsim += 1
            cond = data_defects["atoms"]["type"] == 0
            sia, vac = data_defects["atoms"][~cond], data_defects["atoms"][cond]
            iaclusters, ioclusters = clusterize(sia, cutoff_sia)
            vaclusters, voclusters = clusterize(vac, cutoff_vac)
            aclusters = np.concatenate((iaclusters, vaclusters))
            oclusters = np.concatenate((ioclusters, voclusters))

            data_aclusters = defaultdict(None)
            data_aclusters["timestep"] = data_defects["timestep"]
            data_aclusters["time"] = data_defects["time"]
            data_aclusters["natoms"] = len(aclusters)
            data_aclusters["boundary"] = data_defects["boundary"]
            data_aclusters["xlo"] = data_defects["xlo"]
            data_aclusters["xhi"] = data_defects["xhi"]
            data_aclusters["ylo"] = data_defects["ylo"]
            data_aclusters["yhi"] = data_defects["yhi"]
            data_aclusters["zlo"] = data_defects["zlo"]
            data_aclusters["zhi"] = data_defects["zhi"]
            data_aclusters["atoms"] = aclusters
            awriter.write(data_aclusters)

            data_oclusters = defaultdict(None)
            data_oclusters["timestep"] = data_defects["timestep"]
            data_oclusters["time"] = data_defects["time"]
            data_oclusters["natoms"] = len(oclusters)
            data_oclusters["boundary"] = data_defects["boundary"]
            data_oclusters["xlo"] = data_defects["xlo"]
            data_oclusters["xhi"] = data_defects["xhi"]
            data_oclusters["ylo"] = data_defects["ylo"]
            data_oclusters["yhi"] = data_defects["yhi"]
            data_oclusters["zlo"] = data_defects["zlo"]
            data_oclusters["zhi"] = data_defects["zhi"]
            data_oclusters["atoms"] = oclusters
            owriter.write(data_oclusters)


def atom_to_object(aclusters: dtypes.Acluster) -> dtypes.Ocluster:
    """Transform atom clusters into object clusters.

    Parameters
    ----------
    aclusters : dtypes.Acluster
        Atomic clusters.

    Returns
    -------
    dtypes.Ocluster
        Object clusters.
    """
    nclusters = np.unique(aclusters["cluster"])
    oclusters = np.empty(nclusters.size, dtype=dtypes.ocluster)
    for i in range(nclusters.size):
        acluster = aclusters[aclusters["cluster"] == nclusters[i]]
        oclusters[i]["x"] = np.mean(acluster["x"])
        oclusters[i]["y"] = np.mean(acluster["y"])
        oclusters[i]["z"] = np.mean(acluster["z"])
        oclusters[i]["type"] = acluster[0]["type"]
        oclusters[i]["size"] = acluster.size
    return oclusters


# region Histograms


def get_clusters_0d(
    path_oclusters: Path,
    path_db: Path,
    scale: float = 1.0,
) -> None:
    """Perform a cluster size histogram for interstitials and vacancies.

    This is useful for cluster analysis and cluster dynamics codes.

    Parameters
    ----------
    path_oclusters : Path
        Path to the input file containing object clusters.
    path_db : Path
        Path to the output SQLite database file.
    scale : float, optional (default=1.0)
        All bin counts are multiplied by this factor. Useful for normalization.
    """

    # Extract data
    isizes, vsizes = np.array([], dtype=np.int64), np.array([], dtype=np.int64)
    reader = LAMMPSReader(path_oclusters)
    for data_oclusters in reader:
        cond = data_oclusters["atoms"]["type"] == 0
        vacancies = data_oclusters["atoms"][cond]
        if vacancies.size > 0:
            vsizes = np.concatenate((vsizes, vacancies["size"]))
        interstitials = data_oclusters["atoms"][~cond]
        if interstitials.size > 0:
            isizes = np.concatenate((isizes, interstitials["size"]))

    # Interstitials
    # Histogram sizes
    max_isize = isizes.max()
    # [1:] because size 0 does not count; max_isize + 1 because np.bincount counts from 0
    isize_histogram = np.bincount(isizes, minlength=max_isize + 1)[1:]

    # Vacancies
    # Histogram sizes
    max_vsize = vsizes.max()
    # [1:] because size 0 does not count; max_vsize + 1 because np.bincount counts from 0
    vsize_histogram = np.bincount(vsizes, minlength=max_vsize + 1)[1:]

    # Save to database
    utils.sqlite.insert_array(
        path_db,
        "clusters_0D",
        interstitials=isize_histogram * scale,
        vacancies=vsize_histogram * scale,
    )


def get_clusters_1d(
    path_oclusters: Path,
    path_db: Path,
    axis: str = "x",
    depth_bins: int = 100,
    depth_offset: float = 0.0,
    min_depth: None | float = None,
    max_depth: None | float = None,
    scale: float = 1.0,
) -> None:
    """Perform a cluster size histogram as a function of depth along a specified axis.

    This is useful for cluster analysis and cluster dynamics codes.

    Parameters
    ----------
    path_oclusters : Path
        Path to the input file containing object clusters.
    path_db : Path
        Path to the output SQLite database file.
    axis : str, optional (default="x")
        Axis along which to measure depth. It can be `"x"`, `"y"`, or `"z"`.
    depth_bins : int, optional (default=100)
        Number of bins for the depth histogram.
    depth_offset : float, optional (default=0.0)
        Offset to add to the depth values.
    min_depth : float, optional (default=None)
        Minimum depth to consider. If `None`, the minimum depth from the data is used.
    max_depth : float, optional (default=None)
        Maximum depth to consider. If `None`, the maximum depth from the data is used.
    scale : float, optional (default=1.0)
        All bin counts are multiplied by this factor. Useful for normalization.
    """

    # Extract data
    isizes, vsizes = np.array([], dtype=np.int64), np.array([], dtype=np.int64)
    idepths, vdepths = np.array([], dtype=np.float64), np.array([], dtype=np.float64)
    reader = LAMMPSReader(path_oclusters)
    for data_oclusters in reader:
        cond = data_oclusters["atoms"]["type"] == 0
        vacancies = data_oclusters["atoms"][cond]
        if vacancies.size > 0:
            vsizes = np.concatenate((vsizes, vacancies["size"]))
            vdepths = np.concatenate((vdepths, vacancies[axis]))
        interstitials = data_oclusters["atoms"][~cond]
        if interstitials.size > 0:
            isizes = np.concatenate((isizes, interstitials["size"]))
            idepths = np.concatenate((idepths, interstitials[axis]))
    vdepths += depth_offset
    idepths += depth_offset

    # Histogram depths
    if min_depth is None:
        min_depth = min(idepths.min(), vdepths.min())
    if max_depth is None:
        max_depth = max(idepths.max(), vdepths.max())
    depth_edges = np.histogram_bin_edges(
        idepths, bins=depth_bins, range=(min_depth, max_depth)
    )

    # Interstitials
    # For each depth bin, count the number of each cluster size
    max_isize = isizes.max()
    isize_histogram = np.zeros((depth_bins, max_isize), dtype=np.float64)
    for i in range(depth_bins):
        cond = (idepths >= depth_edges[i]) & (idepths < depth_edges[i + 1])
        # [1:] because size 0 does not count; max_isize + 1 because np.bincount counts from 0
        isize_histogram[i, :] = np.bincount(isizes[cond], minlength=max_isize + 1)[1:]

    # Vacancies
    # For each depth bin, count the number of each cluster size
    max_vsize = vsizes.max()
    vsize_histogram = np.zeros((depth_bins, max_vsize), dtype=np.float64)
    for i in range(depth_bins):
        cond = (vdepths >= depth_edges[i]) & (vdepths < depth_edges[i + 1])
        # [1:] because size 0 does not count; max_vsize + 1 because np.bincount counts from 0
        vsize_histogram[i, :] = np.bincount(vsizes[cond], minlength=max_vsize + 1)[1:]

    # Save to database
    utils.sqlite.insert_array(
        path_db,
        f"clusters_1D_{axis}",
        depth_edges=depth_edges,
        interstitials=isize_histogram * scale,
        vacancies=vsize_histogram * scale,
    )


def read_clusters_0d(path_db: Path) -> dict[str, npt.NDArray[np.float64]]:
    """Read the 0D cluster size histogram from the database.

    This is useful for cluster analysis and cluster dynamics codes.

    Parameters
    ----------
    path_db : Path
        Path to the SQLite database file.

    Returns
    -------
    dict[str, npt.NDArray[np.float64]]
        A dictionary containing:
        - "interstitials": The histogram of interstitial sizes.
        - "vacancies": The histogram of vacancy sizes.
    """
    data = utils.sqlite.read_array(path_db, "clusters_0D")
    return data


def read_clusters_1d(
    path_db: Path, axis: str = "x"
) -> dict[str, npt.NDArray[np.float64]]:
    """Read the 1D cluster size histogram from the database.

    This is useful for cluster analysis and cluster dynamics codes.

    Parameters
    ----------
    path_db : Path
        Path to the SQLite database file.
    axis : str, optional (default="x")
        Axis along which the histogram was computed. It can be `"x"`, `"y"`,or `"z"`.

    Returns
    -------
    dict[str, npt.NDArray[np.float64]]
        A dictionary containing:
        - "depth_edges": The edges of the depth bins.
        - "interstitials": The histogram of interstitial sizes for each depth bin.
        - "vacancies": The histogram of vacancy sizes for each depth bin.
    """
    data = utils.sqlite.read_array(path_db, f"clusters_1D_{axis}")
    return data


# region Plots


def plot_size_1d(
    db_path: Path,
    path_sias: None | Path = None,
    path_vacs: None | Path = None,
    axis: str = "x",
    depth_offset: float = 0.0,
    transpose: bool = True,
    dpi: int = 300,
) -> None:
    """Plot the depth-size 1D histogram for interstitials and vacancies.

    Note
    ----
    The color bar's label shows "Mean counts", then it assumes that in
    `irradiapy.analysis.clusters.get_clusters_1d` a proper `scale` parameter was used.

    Parameters
    ----------
    db_path : Path
        Path to the SQLite database file containing the cluster data generated by
        `irradiapy.analysis.clusters.get_clusters_1d`.
    path_sias : Path, optional (default=None)
        Path to save the interstitials plot. If `None`, the plot is shown instead of saved.
    path_vacs : Path, optional (default=None)
        Path to save the vacancies plot. If `None`, the plot is shown instead of saved.
    axis : str, optional (default="x")
        Axis along which the histogram was computed. It can be `"x"`, `"y"`, or `"z"`.
    depth_offset : float, optional (default=0.0)
        Offset to add to the depth values.
    transpose : bool, optional (default=True)
        If `True`, the depth is on the x-axis and size on the y-axis. If `False`, the
        axes are swapped.
    dpi : int, optional (default=300)
        Dots per inch.
    """

    def plot(
        depth_edges: npt.NDArray[np.float64],
        histogram: npt.NDArray[np.float64],
        title: str,
        path_plot: Path,
        transpose: bool = True,
    ) -> None:

        min_depth, max_depth = depth_edges[0], depth_edges[-1]
        depth_centers = (depth_edges[:-1] + depth_edges[1:]) / 2.0
        min_size, max_size = 1, histogram.shape[1]
        size_centers = np.arange(min_size, max_size + 1)

        if transpose:
            fig = plt.figure(figsize=(8, 6))
            gs = GridSpec(1, 2, width_ratios=[1, 0.05])

            ax = fig.add_subplot(gs[0, 0])
            ax.set_xlabel(r"Depth ($\mathrm{\AA}$)")
            ax.set_ylabel("Cluster size")
            ax.set_ylim(min_size, max_size)
            ax.set_xlim(min_depth, max_depth)

            im = NonUniformImage(
                ax,
                interpolation="nearest",
                norm="log",
                extent=(min_depth, max_depth, min_size, max_size),
            )
            im.set_data(depth_centers, size_centers, histogram.T)
            ax.add_image(im)

            cax = fig.add_subplot(gs[0, 1])
            cbar = fig.colorbar(ax.images[0], cax=cax, orientation="vertical")
            cbar.set_label("Counts per ion")
            cbar.ax.yaxis.set_ticks_position("right")
            cbar.ax.yaxis.set_label_position("right")

        else:
            fig = plt.figure(figsize=(8, 6))
            gs = GridSpec(1, 2, width_ratios=[1, 0.05])

            ax = fig.add_subplot(gs[0, 0])
            ax.set_ylabel(r"Depth ($\mathrm{\AA}$)")
            ax.set_xlabel("Cluster size")
            ax.set_xlim(min_size, max_size)
            ax.set_ylim(min_depth, max_depth)

            im = NonUniformImage(
                ax,
                interpolation="nearest",
                norm="log",
                extent=(min_size, max_size, min_depth, max_depth),
            )
            im.set_data(size_centers, depth_centers, histogram)
            ax.add_image(im)

            cax = fig.add_subplot(gs[0, 1])
            cbar = fig.colorbar(ax.images[0], cax=cax, orientation="vertical")
            cbar.set_label("Counts per ion")
            cbar.ax.yaxis.set_ticks_position("right")
            cbar.ax.yaxis.set_label_position("right")

        fig.suptitle(title)
        fig.tight_layout()
        if path_plot:
            plt.savefig(path_plot, dpi=dpi)
        else:
            plt.show()
        plt.close(fig)

    data = read_clusters_1d(db_path, axis=axis)
    depth_edges = data["depth_edges"] + depth_offset
    interstitials = data["interstitials"]
    vacancies = data["vacancies"]
    plot(depth_edges, interstitials, "Interstitials", path_sias, transpose)
    plot(depth_edges, vacancies, "Vacancies", path_vacs, transpose)


def plot_clustering_fraction_1d(
    db_path: Path,
    path_plot: None | Path = None,
    axis: str = "x",
    depth_offset: float = 0.0,
    dpi: int = 300,
) -> None:
    """Plot the clustering fraction as a function of depth.

    Note
    ----
    The clustering fraction is defined as the ratio of the number of defects in clusters of size
    greater than 1 to the number of unclustered defects.

    Parameters
    ----------
    db_path : Path
        Path to the SQLite database file containing the cluster data generated by
        `irradiapy.analysis.clusters.get_clusters_1d`.
    path_plot : Path, optional (default=None)
        Path to save the clustering fraction plot. If `None`, the plot is shown instead of saved.
    axis : str, optional (default="x")
        Axis along which the histogram was computed. It can be `"x"`, `"y"`, or `"z"`.
    depth_offset : float, optional (default=0.0)
        Offset to add to the depth values.
    dpi : int, optional (default=300)
        Dots per inch.
    """
    data = read_clusters_1d(db_path, axis=axis)
    depth_edges = data["depth_edges"] + depth_offset
    interstitials = data["interstitials"]
    vacancies = data["vacancies"]

    min_depth, max_depth = depth_edges[0], depth_edges[-1]
    depth_centers = (depth_edges[:-1] + depth_edges[1:]) / 2.0

    min_size, max_size = 1, interstitials.shape[1]
    size_centers = np.arange(min_size, max_size + 1)
    interstitials *= size_centers
    clustering_fraction_sia = np.sum(interstitials[:, 1:], axis=1) / interstitials[:, 0]

    min_size, max_size = 1, vacancies.shape[1]
    size_centers = np.arange(min_size, max_size + 1)
    vacancies *= size_centers
    clustering_fraction_vac = np.sum(vacancies[:, 1:], axis=1) / vacancies[:, 0]

    fig = plt.figure(figsize=(8, 6))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_ylabel("Clustering fraction")
    ax.plot(
        depth_centers,
        clustering_fraction_sia,
        marker="o",
        linestyle="none",
        label="Interstitials",
    )
    ax.plot(
        depth_centers,
        clustering_fraction_vac,
        marker="o",
        linestyle="none",
        label="Vacancies",
    )
    ax.set_ylim(0, 1)
    ax.set_xlim(min_depth, max_depth)
    ax.legend()
    fig.suptitle("Clustering fraction")
    fig.tight_layout()
    if path_plot:
        plt.savefig(path_plot, dpi=dpi)
    else:
        plt.show()
    plt.close(fig)


def plot_mddb_cluster_size(
    target_dir: Path,
    sia_cutoff: float,
    vac_cutoff: float,
    rel: bool,
    sia_bin_width: int = 10,
    vac_bin_width: int = 10,
    path_sias: None | Path = None,
    path_vacs: None | Path = None,
    max_sia: None | int = None,
    max_vac: None | int = None,
    vmin: float = 1e-5,
    dpi: int = 300,
) -> None:
    """Plot the cluster size distribution from a molecular dynamics database.

    Parameters
    ----------
    target_dir : Path
        Directory containing the MD database.
    sia_cutoff : float
        Cutoff distance for interstitial cluster identification.
    vac_cutoff : float
        Cutoff distance for vacancy cluster identification.
    rel : bool
        If `True`, each size bin is the number of clusters of that size divided by the total number
        of clusters for the given energy. If `False`, each size bin is the number of clusters of
        that size divided by the total number of simulations of the corresponding energy.
    sia_bin_width : int, optional (default=10)
        Bin size for interstitial histogram.
    vac_bin_width : int, optional (default=10)
        Bin size for vacancy histogram.
    path_sias : Path, optional (default=None)
        Path to save the interstitials size distribution plot. If `None`, the plot is shown instead
        of saved.
    path_vacs : Path, optional (default=None)
        Path to save the vacancies size distribution plot. If `None`, the plot is shown instead
        of saved.
    max_sia : int, optional (default=None)
        Sets an upper limit for interstitial cluster sizes. It has to be greater than the maximum
        interstitial cluster size in the data. If `None`, it is determined from the data. Useful to
        compare multiple databases with different maximum sizes.
    max_vac : int, optional (default=None)
        Sets an upper limit for vacancy cluster sizes. It has to be greater than the maximum
        vacancy cluster size in the data. If `None`, it is determined from the data. Useful to
        compare multiple databases with different maximum sizes.
    vmin : float, optional (default=1e-5)
        Minimum value for the color scale if `rel` is `True`.
    dpi : int, optional (default=300)
        Dots per inch.
    """

    # Extract data from xyz files
    sizes_all_raw = defaultdict(defaultdict)
    nfiles = defaultdict(int)
    for energy_dir in target_dir.glob("*"):
        if not energy_dir.is_dir():
            continue
        energy = int(energy_dir.name)
        sizes_all_raw[energy] = {
            "vacs": np.empty(0, dtype=np.int64),
            "sias": np.empty(0, dtype=np.int64),
        }
        for path_defects in energy_dir.glob("*.xyz"):
            nfiles[energy] += 1
            data_defects = utils.io.get_last_lammps_dump(path_defects)

            cond = data_defects["atoms"]["type"] == 0
            vacs = data_defects["atoms"][cond]
            sias = data_defects["atoms"][~cond]

            _, vac_oclusters = clusterize(vacs, vac_cutoff)
            _, sia_oclusters = clusterize(sias, sia_cutoff)

            sizes_all_raw[energy]["vacs"] = np.append(
                sizes_all_raw[energy]["vacs"],
                vac_oclusters["size"],
            )
            sizes_all_raw[energy]["sias"] = np.append(
                sizes_all_raw[energy]["sias"],
                sia_oclusters["size"],
            )

    # Maximum cluster sizes
    if max_sia is None or max_vac is None:
        max_sia = max_vac = 0
        for energy, siasvacs in sizes_all_raw.items():
            max_sia = max(max_sia, siasvacs["sias"].max())
            max_vac = max(max_vac, siasvacs["vacs"].max())

    # Bincount sizes (exclude size 0)
    sizes_all_bincount = defaultdict(defaultdict)
    for energy, sizes in sizes_all_raw.items():
        sizes_all_bincount[energy]["vacs"] = np.bincount(
            sizes["vacs"], minlength=max_vac
        )[1:]
        sizes_all_bincount[energy]["sias"] = np.bincount(
            sizes["sias"], minlength=max_sia
        )[1:]

    # Sort energies
    epkas = np.array(
        [int(path.name) for path in target_dir.glob("*") if path.is_dir()],
        dtype=np.int64,
    )
    epkas.sort()
    # Energy binning
    nepka_bins = len(epkas)

    # Round to closest highest multiple of bin_width starting from 10,
    # where binning changes
    # if bin_width = 10, then 83 > 90
    # if bin_width = 10, then 57 > 60
    max_sia = 10 + int(np.ceil((max_sia - 10) / sia_bin_width) * sia_bin_width)
    max_vac = 10 + int(np.ceil((max_vac - 10) / vac_bin_width) * vac_bin_width)

    # Histogram bin edges
    sia_edges = np.concatenate(
        (
            np.arange(0.5, 10.5),
            np.arange(10.5, max_sia + sia_bin_width + 0.5, sia_bin_width),
        )
    )
    vac_edges = np.concatenate(
        (
            np.arange(0.5, 10.5),
            np.arange(10.5, max_vac + vac_bin_width + 0.5, vac_bin_width),
        )
    )

    # row: energy, column: size
    ihist = np.empty((nepka_bins, len(sia_edges) - 1), dtype=np.float64)
    vhist = np.empty((nepka_bins, len(vac_edges) - 1), dtype=np.float64)
    for i, epka in enumerate(epkas):
        siasvacs = sizes_all_raw[epka]
        if rel:
            hist, _ = np.histogram(siasvacs["sias"], bins=sia_edges, density=True)
            ihist[i] = hist
            hist, _ = np.histogram(siasvacs["vacs"], bins=vac_edges, density=True)
            vhist[i] = hist
        else:
            hist, _ = np.histogram(siasvacs["sias"], bins=sia_edges, density=False)
            ihist[i] = hist / nfiles[epka]
            hist, _ = np.histogram(siasvacs["vacs"], bins=vac_edges, density=False)
            vhist[i] = hist / nfiles[epka]

    def plot(
        hist: npt.NDArray[np.float64],
        edges: npt.NDArray[np.float64],
        max_size: int,
        bin_width: int,
        title: str,
        path: None | Path = None,
    ) -> None:

        hist = np.ma.masked_where(hist == 0, hist, copy=False)

        fig = plt.figure(figsize=(8, 8))
        gs = GridSpec(1, 2, height_ratios=[1], width_ratios=[1, 0.05])
        ax = fig.add_subplot(gs[0, 0])
        ax.set_xlabel("Energy (keV)")
        ax.set_ylabel("Cluster size")

        if rel:
            norm = mcolors.LogNorm(vmin=vmin, vmax=1.0)
        else:
            norm = mcolors.LogNorm(vmin=hist.min(), vmax=hist.max())
        im = ax.imshow(
            hist.T,
            origin="lower",
            aspect="auto",
            norm=norm,
            cmap="viridis",
            extent=[0, nepka_bins, 0, hist.shape[1]],
        )

        xticks = np.arange(nepka_bins) + 0.5
        xlabels = [f"{epka/1000.0:<.1f}" for epka in epkas]
        ax.set_xticks(xticks)
        ax.set_xticklabels(xlabels)

        # Size labels: 1, 2, ..., 10, 11-20, ...
        if bin_width == 1:
            ylabels = [str(i) for i in range(1, max_size + 1)]
        else:
            ylabels = [str(i) for i in range(1, 11)] + [
                f"{i}-{i+bin_width-1}" for i in range(11, max_size, bin_width)
            ]
        yticks = np.arange(len(edges) - 1) + 0.5
        ax.set_yticks(yticks)
        ax.set_yticklabels(ylabels)

        cax = fig.add_subplot(gs[0, 1])
        fig.colorbar(im, cax=cax, label="% of counts per energy")

        fig.suptitle(title)
        fig.tight_layout()
        if path:
            plt.savefig(path, dpi=dpi)
        else:
            plt.show()
        plt.close(fig)

    plot(ihist, sia_edges, max_sia, sia_bin_width, "Interstitials", path=path_sias)
    plot(vhist, vac_edges, max_vac, vac_bin_width, "Vacancies", path=path_vacs)



"""Class to read LAMMPS log files."""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Generator

import numpy as np
from numpy import typing as npt


@dataclass
class LAMMPSLogReader:
    """Class to read LAMMPS log files.

    Parameters
    ----------
    path_log : Path
        The path to the LAMMPS log file.

    Yields
    -------
    Generator[dict[str, Any], None, None]
        A dictionary containing the thermo data.

    Note
    ----
    This generator yields a dictionary with a single key for the thermo data,
    this ensures compatibility if this reader is extended to read more data in the future.
    """

    path_log: Path
    data: dict = field(default_factory=lambda: {"thermo": None}, init=False)
    thermo: npt.NDArray[np.float64] = field(
        default_factory=lambda: np.empty(0, dtype=np.dtype([])), init=False
    )

    def __iter__(self) -> Generator[dict[str, Any], None, None]:
        """Reads a LAMMPS log file.

        Parameters
        ----------
        path_log : Path
            The path to the log file.

        Returns
        -------
        npt.NDArray
            The data as a structured array.
        """
        self.__reset()
        expecting_header = False  # Next line will be a header
        expecting_data = False  # Next lines will be data
        for line in open(self.path_log, "r", encoding="utf-8"):
            if line.startswith("Per MPI "):
                expecting_header = True
                self.__reset()
                continue
            elif line.startswith("Loop time") or line.startswith("Fix halt"):
                if self.thermo.size > 0:
                    self.__fill()
                    yield self.data
                expecting_header = False
                expecting_data = False
                self.__reset()
                continue
            if expecting_header:
                items = line.split()
                types = [np.float64] * len(items)
                if "Step" in items:
                    types[items.index("Step")] = np.int64
                dtype = np.dtype(list(zip(items, types)))
                self.thermo = np.empty(0, dtype=dtype)
                expecting_header = False
                expecting_data = True
                continue
            if expecting_data:
                if line.startswith("WARNING"):
                    continue
                row = np.array(tuple(map(float, line.split())), dtype=dtype)
                self.thermo = np.append(self.thermo, row)

    def __reset(self) -> None:
        """Reset the internal state of the reader."""
        self.thermo = np.empty(0, dtype=np.dtype([]))
        self.data = {"thermo": self.thermo}

    def __fill(self) -> None:
        """Fill the data dictionary with data."""
        self.data["thermo"] = self.thermo



"""Ag material."""

from irradiapy.materials.material import Material
from irradiapy.srim.target import element as srim_element

Ag = Material(
    atomic_number=47,
    mass_number=107.87,
    srim_element=srim_element.Element(
        symbol="Ag",
        atomic_number=47,
        atomic_mass=107.8682,
        e_d=39.0,
        e_l=4.0,
        e_s=2.97,
        density=10.49,
    ),
)



"""This subpackage provides classes for reading and writing various file formats."""

from .bzip2lammpsreader import BZIP2LAMMPSReader
from .bzip2lammpswriter import BZIP2LAMMPSWriter
from .lammpslogreader import LAMMPSLogReader
from .lammpsreader import LAMMPSReader
from .lammpsreadermpi import LAMMPSReaderMPI
from .lammpswriter import LAMMPSWriter
from .lammpswritermpi import LAMMPSWriterMPI
from .xyzreader import XYZReader
from .xyzwriter import XYZWriter



"""This subpackage provides a Python interface for SRIM for simulating collisional cascades."""

from irradiapy.srim import analysis
from irradiapy.srim.srimdb import SRIMDB
from irradiapy.srim.target import Element, Layer, Target



"""This module contains the `Element` class."""

from dataclasses import dataclass


@dataclass
class Element:
    """Class for defining an element in SRIM simulations.

    Attributes
    ----------
    symbol : str
        The symbol of the element.
    atomic_number : int
        The atomic number of the element.
    atomic_mass : float
        The atomic mass of the element in atomic mass units.
    e_d : float
        The displacement energy of the element in eV.
    e_l : float
        The lattice binding energy of the element in eV.
    e_s : float
        The surface binding energy of the element in eV.
    density : float, optional
        The density of the element in g/cm^3. Only to be used by predefined
        materials to simplify the definition of the layers in the target by the
        user.
    """

    symbol: str
    atomic_number: int
    atomic_mass: float
    e_d: float
    e_l: float
    e_s: float
    density: None | float = None



"""This module contains math utilities for the irradiapy package."""

# pylint: disable=unbalanced-tuple-unpacking

from typing import Callable

import numpy as np
from numpy import typing as npt
from scipy.optimize import curve_fit


def repeated_prime_factors(n: int) -> list[int]:
    """Return the prime factors of n (with repetition).

    Parameters
    ----------
    n : int
        The number to factorize.

    Returns
    -------
    list[int]
        List of prime factors of n, including repetitions.
    """
    facs = []
    # Factor out 2s
    while n % 2 == 0:
        facs.append(2)
        n //= 2
    # Factor odd primes up to sqrt(n)
    f = 3
    while f * f <= n:
        while n % f == 0:
            facs.append(f)
            n //= f
        f += 2
    if n > 1:
        facs.append(n)
    return facs


# region Lorentzian


def lorentzian(
    xs: npt.NDArray[np.float64],
    x_peak: float,
    linewidth: float,
    amplitude: float,
    asymmetry: float,
) -> float | npt.NDArray[np.float64]:
    """Evaluate a Lorentzian function.

    Parameters
    ----------
    xs : npt.NDArray[np.float64]
        Where to evaluate the function.
    x_peak : float
        Position with maximum value.
    linewidth : float
        Linewidth.
    amplitude : float
        Maximum amplitude.
    asymmetry : float
        Asymmetry.

    Returns
    -------
    float | npt.NDArray[np.float64]
        Evaluated Lorentzian function.

    References
    ----------
    See https://doi.org/10.1016/j.nimb.2021.05.014
    """
    delta_x = xs - x_peak
    linewidth_sq = linewidth**2
    exp_term = np.exp(asymmetry * delta_x)
    alpha = (1.0 + exp_term) ** 2
    alpha_quarter = alpha / 4.0
    exponent = -alpha_quarter * delta_x**2 / (2.0 * linewidth_sq)
    return amplitude * alpha_quarter * np.exp(exponent)


def fit_lorentzian(
    xs: npt.NDArray[np.float64],
    ys: npt.NDArray[np.float64],
    p0: None | npt.NDArray[np.float64] = None,
    asymmetry: float = 1.0,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]],
]:
    """Fit data to a Lorentzian function.

    Parameters
    ----------
    xs : npt.NDArray[np.float64]
        X values where the function is evaluated.
    ys : npt.NDArray[np.float64]
        Y values at the given xs.
    p0 : npt.NDArray[np.float64], optional (default=None)
        Initial guess of fit parameters. If None, a guess is generated.
    asymmetry : float, optional (default=1.0)
        Bound for the asymmetry fit parameter. Fit will be done in (-asymmetry, asymmetry).

    Returns
    -------
    popt : npt.NDArray[np.float64]
        Optimal values for the parameters.
    pcov : npt.NDArray[np.float64]
        Covariance of popt.
    fit_function : Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]
        Function that evaluates the fitted Lorentzian.
    """
    if p0 is None:
        peak_index = np.argmax(ys)
        x_start = xs[0]
        x_end = xs[-1]
        sigma_guess = 0.5 * (x_end - x_start)
        p0 = np.array(
            [
                xs[peak_index],
                sigma_guess,
                ys[peak_index],
                0.0,
            ]
        )
    x_start = xs[0]
    x_end = xs[-1]
    x_sum = x_start + x_end
    popt, pcov = curve_fit(
        lorentzian,
        xs,
        ys,
        p0=p0,
        bounds=(
            [x_start, 0.0, ys.min(), -asymmetry],
            [x_end, x_sum, ys.max(), asymmetry],
        ),
    )

    def fit_function(xs_fit: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        return lorentzian(xs_fit, *popt)

    return popt, pcov, fit_function


# region Gaussian


def gaussian(
    xs: npt.NDArray[np.float64],
    x_peak: float,
    linewidth: float,
    amplitude: float,
    asymmetry: float,
) -> float | npt.NDArray[np.float64]:
    """Evaluate a Gaussian function.

    Parameters
    ----------
    xs : npt.NDArray[np.float64]
        Where to evaluate the function.
    x_peak : float
        Position with maximum value.
    linewidth : float
        Linewidth.
    amplitude : float
        Maximum amplitude.
    asymmetry : float
        Asymmetry.

    Returns
    -------
    float | npt.NDArray[np.float64]
        Evaluated Gaussian function.

    References
    ----------
    See https://doi.org/10.1016/j.nimb.2021.05.014
    """
    delta_x = xs - x_peak
    linewidth_sq = linewidth**2
    exp_term = np.exp(asymmetry * delta_x)
    alpha = (1.0 + exp_term) ** 2
    alpha_quarter = alpha / 4.0
    exponent = -alpha_quarter * delta_x**2 / (2.0 * linewidth_sq)
    return amplitude * alpha_quarter * np.exp(exponent)


def fit_gaussian(
    xs: npt.NDArray[np.float64],
    ys: npt.NDArray[np.float64],
    p0: None | npt.NDArray[np.float64] = None,
    asymmetry: float = 1.0,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]],
]:
    """Fit data to a Gaussian function.

    Parameters
    ----------
    xs : npt.NDArray[np.float64]
        X values where the function is evaluated.
    ys : npt.NDArray[np.float64]
        Y values at the given xs.
    p0 : npt.NDArray[np.float64], optional (default=None)
        Initial guess of fit parameters. If None, a guess is generated.
    asymmetry : float, optional (default=1.0)
        Bound for the asymmetry fit parameter. Fit will be done in (-asymmetry, asymmetry).

    Returns
    -------
    popt : npt.NDArray[np.float64]
        Optimal values for the parameters.
    pcov : npt.NDArray[np.float64]
        Covariance of popt.
    fit_function : Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]
        Function that evaluates the fitted Gaussian.
    """
    if p0 is None:
        peak_index = np.argmax(ys)
        x_start = xs[0]
        x_end = xs[-1]
        sigma_guess = 0.5 * (x_end - x_start)
        p0 = np.array(
            [
                xs[peak_index],
                sigma_guess,
                ys[peak_index],
                0.0,
            ]
        )
    x_start = xs[0]
    x_end = xs[-1]
    x_sum = x_start + x_end
    popt, pcov = curve_fit(
        gaussian,
        xs,
        ys,
        p0=p0,
        bounds=(
            [x_start, 0.0, ys.min(), -asymmetry],
            [x_end, x_sum, ys.max(), asymmetry],
        ),
    )

    def fit_function(xs_fit: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        return gaussian(xs_fit, *popt)

    return popt, pcov, fit_function


# region Power law


def scaling_law(
    x: npt.NDArray[np.float64], a: float, s: float
) -> npt.NDArray[np.float64]:
    """Evaluate the scaling law function a / x**s.

    Parameters
    ----------
    x : npt.NDArray[np.float64]
        Input values.
    a : float
        Prefactor.
    s : float
        Exponent.

    Returns
    -------
    npt.NDArray[np.float64]
        Evaluated scaling law.
    """
    return a / x**s


def fit_scaling_law(
    centers: npt.NDArray[np.float64], counts: npt.NDArray[np.float64]
) -> tuple[float, float, Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]]:
    """Fit a scaling law to the given histogram data.

    Parameters
    ----------
    centers : npt.NDArray[np.float64]
        The centers of the bins.
    counts : npt.NDArray[np.float64]
        The values of the histogram.

    Returns
    -------
    tuple
        A tuple containing: the prefactor of the scaling law, the exponent of the scaling law,
        and the scaling law function.
    """
    popt, _ = curve_fit(lambda x, a, b: a + b * x, np.log10(centers), np.log10(counts))
    a, s = popt
    a, s = 10.0**a, -s

    def fit_function(x: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        return scaling_law(x, a, s)

    return a, s, fit_function



"""Subpackage for materials."""

from irradiapy.materials.ag import Ag
from irradiapy.materials.cr import Cr
from irradiapy.materials.cu import Cu
from irradiapy.materials.fe import Fe
from irradiapy.materials.material import Material
from irradiapy.materials.o import O
from irradiapy.materials.w import W

MATERIALS_BY_SYMBOL = {
    "Ag": Ag,
    "Cr": Cr,
    "Cu": Cu,
    "Fe": Fe,
    "O": O,
    "W": W,
}
MATERIALS_BY_ATOMIC_NUMBER = {
    8: O,
    24: Cr,
    26: Fe,
    29: Cu,
    47: Ag,
    74: W,
}



"""Cr material."""

from irradiapy.materials.material import Material
from irradiapy.srim.target import element as srim_element

Cr = Material(
    atomic_number=24,
    mass_number=51.996,
    srim_element=srim_element.Element(
        symbol="Cr",
        atomic_number=24,
        atomic_mass=51.9961,
        e_d=40.0,
        e_l=7.8,
        e_s=13.2,
        density=7.19,
    ),
)



"""sqlite3 utilities."""

import io
import sqlite3
from pathlib import Path

import numpy as np
from numpy import typing as npt


def insert_array(path_db: Path, name: str, **kargs) -> None:
    """Insert or update an array in the arrays table.

    Parameters
    ----------
    path_db : Path
        Path to the SQLite database file.
    name : str
        Name of the array to insert or update.
    kargs : dict
        Keyword arguments representing the array data.
    """
    buffer = io.BytesIO()
    np.savez(buffer, **kargs)
    blob = buffer.getvalue()
    buffer.close()
    conn = sqlite3.connect(path_db)
    cur = conn.cursor()
    cur.execute("CREATE TABLE IF NOT EXISTS arrays (name TEXT UNIQUE, data BLOB)")
    cur.execute(
        "INSERT OR REPLACE INTO arrays (name, data) VALUES (?, ?)",
        (name, blob),
    )
    conn.commit()
    cur.close()
    conn.close()


def read_array(path_db: Path, name: str) -> dict[str, npt.NDArray]:
    """Read an array from the arrays table.

    Parameters
    ----------
    path_db : Path
        Path to the SQLite database file.
    name : str
        Name of the array to read.

    Returns
    -------
    dict[str, npt.NDArray]
        Dictionary containing the array data.
    """
    conn = sqlite3.connect(path_db)
    cur = conn.cursor()
    cur.execute("SELECT data FROM arrays WHERE name = ?", (name,))
    row = cur.fetchone()
    cur.close()
    conn.close()

    if row is None:
        raise ValueError(f"Array '{name}' not found in the database.")

    buffer = io.BytesIO(row[0])
    data = np.load(buffer, allow_pickle=False)
    data = dict(data.items())
    buffer.close()

    return data


def delete_array(path_db: Path, name: str) -> None:
    """Delete an array from the arrays table.

    Parameters
    ----------
    path_db : Path
        Path to the SQLite database file.
    name : str
        Name of the array to delete.
    """
    conn = sqlite3.connect(path_db)
    cur = conn.cursor()
    cur.execute("DELETE FROM arrays WHERE name = ?", (name,))
    conn.commit()
    cur.close()
    conn.close()


def delete_all_arrays(path_db: Path) -> None:
    """Drop the arrays table from the database.

    Parameters
    ----------
    path_db : Path
        Path to the SQLite database file.
    """
    conn = sqlite3.connect(path_db)
    cur = conn.cursor()
    cur.execute("DROP TABLE IF EXISTS arrays")
    conn.commit()
    cur.close()
    conn.close()



"""This module contains the `Material` class."""

from dataclasses import dataclass
from enum import Enum, auto

import numpy as np
from numpy import typing as npt

# pylint: disable=unused-import
from irradiapy.srim.target import element as srim_element

# pylint: enable=unused-import


@dataclass
class Material:
    """Class for storing parameters of a material.

    Parameters
    ----------
    atomic_number : int
        Atomic number.
    mass_number : float
        Mass number (atomic mass units).
    a0 : float, optional (default=None)
        Lattice parameter (Å).
    cutoff_sia : float, optional (default=None)
        Cutoff distance for interstitial clusters detection (Å).
    cutoff_vac : float, optional (default=None)
        Cutoff distance for vacancy clusters detection (Å).
    dist_fp : float, optional (default=None)
        Frenkel pair distance (Å).
    density : float, optional (default=None)
        Atomic density (atoms/Å³).
    ed_min : float, optional (default=None)
        Minimum displacement energy (eV).
    ed_avr : float, optional (default=None)
        Average displacement energy (eV).
    b_arc : float, optional (default=None)
        'b' parameter of the arc-dpa fit.
    c_arc : float, optional (default=None)
        'c' parameter of the arc-dpa fit.
    srim_element : srim_element.Element, optional (default=None)
        SRIM element object. Check this reference for recommended values:
        https://doi.org/10.1016/j.nimb.2021.06.018.
    """

    atomic_number: int
    mass_number: float
    a0: None | float = None
    cutoff_sia: None | float = None
    cutoff_vac: None | float = None
    dist_fp: None | float = None
    density: None | float = None
    ed_min: None | float = None
    ed_avr: None | float = None
    b_arc: None | float = None
    c_arc: None | float = None
    srim_element: "None | srim_element.Element" = None

    # region Damage energy

    class TdamMode(Enum):
        """Enumeration of damage energy calculation modes."""

        SRIM = auto()
        LINDHARD = auto()

    def epka_to_tdam(
        self, mat_pka: "Material", epka: float, mode: TdamMode = TdamMode.SRIM
    ) -> float:
        """Convert PKA energy to damage energy.

        Parameters
        ----------
        mat_pka : Material
            Material of the PKA.
        epka : float
            PKA energy, in eV.
        mode : Material.TdamMode
            Mode for damage energy calculation. Can be either `Material.TdamMode.SRIM` or
            `Material.TdamMode.LINDHARD`.

        Returns
        -------
        float
            Damage energy (eV).
        """
        if mode == Material.TdamMode.SRIM:
            if self.atomic_number == mat_pka.atomic_number == 26:  # Fe
                #  SRIM Quick-Calculation, D1
                return 699e-3 * epka - 460e-9 * np.square(epka)
            if self.atomic_number == mat_pka.atomic_number == 74:  # W
                # SRIM Quick-Calculation, D1
                return 752e-3 * epka - 216e-9 * np.square(epka)
        else:  # mode == Material.TdamMode.LINDHARD:
            return self.epka_to_tdam_lindhard(mat_pka, epka)
        # else:
        #     raise ValueError("Invalid damage energy calculation mode.")

    def epka_to_tdam_lindhard(self, mat_pka: "Material", epka: float) -> float:
        """Convert PKA energy to damage energy using the Lindhard equation.

        Parameters
        ----------
        mat_pka : Material
            Material of the PKA.
        epka : float
            PKA energy, in eV.

        Returns
        -------
        float
            Damage energy, in eV.
        """
        a0 = 0.529177e-10  # m, Bohr radius
        e2 = 1.4e-9  # eV2 m s, squared unit charge for Lindhard expression
        a = (
            (9.0 * np.pi**2 / 128.0) ** (1.0 / 3.0)
            * a0
            / (mat_pka.atomic_number ** (2.0 / 3.0) + self.atomic_number ** (2.0 / 3.0))
            ** 0.5
        )
        redu = (
            (self.mass_number * epka)
            / (mat_pka.mass_number + self.mass_number)
            * a
            / (mat_pka.atomic_number * self.atomic_number * e2)
        )
        k = (
            0.1337
            * mat_pka.atomic_number ** (1.0 / 6.0)
            * (mat_pka.atomic_number / mat_pka.mass_number) ** 0.5
        )
        g = 3.4008 * redu ** (1.0 / 6.0) + 0.40244 * redu ** (3.0 / 4.0) + redu
        return epka / (1.0 + k * g)

    # region dpa calculation

    class DpaMode(Enum):
        """Enumeration of dpa calculation modes.

        References
        ----------
        NRT : https://doi.org/10.1016/0029-5493(75)90035-7
        ARC : https://doi.org/10.1038/s41467-018-03415-5
        FERARC : https://doi.org/10.1103/PhysRevMaterials.5.073602
        """

        NRT = auto()
        ARC = auto()
        FERARC = auto()

    def tdam_to_dpa(
        self,
        tdam: float | npt.NDArray[np.float64],
        mode: DpaMode = DpaMode.FERARC,
    ) -> int | npt.NDArray[np.float64]:
        """Convert damage energy to dpa.

        Parameters
        ----------
        tdam : float | npt.NDArray[np.float64]
            Damage energy, in eV.
        mode : Material.DpaMode, optional
            Mode for dpa calculation. Can be one of `Material.DpaMode.NRT`, `Material.DpaMode.ARC`,
            or `Material.DpaMode.FERARC`.
        Returns
        -------
        float | npt.NDArray[np.float64]
            Number of Frenkel pairs predicted by the specified dpa mode.
        """
        if mode == Material.DpaMode.FERARC:
            return self.calc_fer_arc_dpa(tdam)
        elif mode == Material.DpaMode.ARC:
            return self.calc_arc_dpa(tdam)
        else:
            return self.calc_nrt_dpa(tdam)

    def calc_nrt_dpa(
        self,
        tdam: float | npt.NDArray[np.float64],
    ) -> float | npt.NDArray[np.float64]:
        """Calculate the NRT-dpa for the given damage energy.

        Parameters
        ----------
        tdam : int | float | numpy.ndarray
            Damage energy in electron volts.

        Returns
        -------
        int | numpy.ndarray
            Number of Frenkel pairs predicted by NRT-dpa.
        """
        min_threshold = self.ed_avr
        max_threshold = 2.5 * self.ed_avr

        def scaling_func(x):
            return 0.4 * x / self.ed_avr

        if isinstance(tdam, (float, int)):
            if tdam < min_threshold:
                return 0.0
            elif tdam > max_threshold:
                return scaling_func(tdam)
            else:
                return 1.0
        elif isinstance(tdam, np.ndarray) and np.issubdtype(tdam.dtype, np.number):
            return self.__apply_dpa_thresholds(
                tdam, min_threshold, max_threshold, scaling_func
            )
        else:
            raise TypeError("tdam must be a number or numpy array of numbers")

    def calc_arc_dpa(
        self,
        tdam: float | npt.NDArray[np.float64],
    ) -> float | npt.NDArray[np.float64]:
        """Calculate the arc-dpa for the given damage energy in eV.

        Parameters
        ----------
        tdam : float | npt.NDArray[np.float64]
            Damage energy in electron volts.

        Returns
        -------
        float | npt.NDArray[np.float64]
            Number of Frenkel pairs predicted by arc-dpa.
        """
        min_threshold = self.ed_avr
        max_threshold = 2.5 * self.ed_avr

        def scaling_func(x):
            return 0.4 * x / self.ed_avr

        def efficiency_func(x):
            return (1.0 - self.c_arc) / (
                max_threshold**self.b_arc
            ) * x**self.b_arc + self.c_arc

        if isinstance(tdam, (float, int)):
            if tdam < min_threshold:
                return 0.0
            elif tdam > max_threshold:
                eff = efficiency_func(tdam)
                return scaling_func(tdam) * eff
            else:
                return 1.0
        elif isinstance(tdam, np.ndarray):
            return self.__apply_dpa_thresholds(
                tdam, min_threshold, max_threshold, scaling_func, efficiency_func
            )
        else:
            raise TypeError("tdam must be a number or numpy array of numbers")

    def calc_fer_arc_dpa(
        self,
        tdam: float | npt.NDArray[np.float64],
    ) -> float | npt.NDArray[np.float64]:
        """Calculate the fer-arc-dpa for the given damage energy.

        Parameters
        ----------
        tdam : float | npt.NDArray[np.float64]
            Damage energy, in eV.

        Returns
        -------
        float | npt.NDArray[np.float64]
            Number of Frenkel pairs predicted by modified arc-dpa.
        """
        min_threshold = self.ed_min
        max_threshold = 2.5 * self.ed_avr

        def scaling_func(x):
            return 0.4 * x / self.ed_avr

        def efficiency_func(x):
            return (1.0 - self.c_arc) / (
                max_threshold**self.b_arc
            ) * x**self.b_arc + self.c_arc

        if isinstance(tdam, (float, int)):
            if tdam < min_threshold:
                return 0.0
            elif tdam > max_threshold:
                eff = efficiency_func(tdam)
                return scaling_func(tdam) * eff
            else:
                return scaling_func(tdam)
        elif isinstance(tdam, np.ndarray):
            return self.__apply_dpa_thresholds(
                tdam,
                min_threshold,
                max_threshold,
                scaling_func,
                efficiency_func,
                scaling_func,
            )
        else:
            raise TypeError("tdam must be a number or numpy array of numbers")

    @staticmethod
    def __apply_dpa_thresholds(
        tdam: npt.NDArray[np.float64],
        min_threshold: float,
        max_threshold: float,
        scaling_func: callable,
        efficiency_func: callable = None,
        middle_func: callable = None,
    ) -> npt.NDArray[np.float64]:
        """Apply dpa thresholds and scaling/efficiency functions.

        Parameters
        ----------
        tdam : npt.NDArray[np.float64]
            Damage energy array.
        min_threshold : float
            Minimum threshold for dpa.
        max_threshold : float
            Maximum threshold for dpa.
        scaling_func : callable
            Function to scale damage energy.
        efficiency_func : callable, optional (default=None)
            Efficiency function for high energies.
        middle_func : callable, optional (default=None)
            Function for values between thresholds.

        Returns
        -------
        npt.NDArray[np.float64]
            Array of dpa values.
        """
        result = np.ones_like(tdam, dtype=np.float64)
        below_mask = tdam < min_threshold
        above_mask = tdam > max_threshold
        result[below_mask] = 0
        if middle_func is not None:
            middle_mask = (~below_mask) & (~above_mask)
            result[middle_mask] = middle_func(tdam[middle_mask])
        # else: keep as 1
        if efficiency_func:
            result[above_mask] = scaling_func(tdam[above_mask]) * efficiency_func(
                tdam[above_mask]
            )
        else:
            result[above_mask] = scaling_func(tdam[above_mask])
        return result



"""Utility functions for I/O operations."""

import bz2
from collections import defaultdict, deque
from pathlib import Path

from irradiapy.io import LAMMPSReader


def compress_file_bz2(
    input_path: str, output_path: str, compresslevel: int = 9
) -> None:
    """Compress a file using bzip2.

    Parameters
    ----------
    input_path : str
        Path to the input file to be compressed.
    output_path : str
        Path where the compressed file will be saved.
    compresslevel : int, optional (default=9)
        Compression level for bzip2.
    """
    with (
        open(input_path, "rb") as f_in,
        bz2.open(output_path, "wb", compresslevel=compresslevel) as f_out,
    ):
        for chunk in iter(lambda: f_in.read(1024 * 1024), b""):
            f_out.write(chunk)


def decompress_file_bz2(input_path: str, output_path: str) -> None:
    """Decompress a bzip2-compressed file.

    Parameters
    ----------
    input_path : str
        Path to the input .bz2 file.
    output_path : str
        Path to the output decompressed file.
    """
    with bz2.open(input_path, "rb") as f_in, open(output_path, "wb") as f_out:
        for chunk in iter(lambda: f_in.read(1024 * 1024), b""):
            f_out.write(chunk)


def get_last_lammps_dump(path: Path) -> defaultdict:
    """Get the last snaptshot from a LAMMPS dump file.

    Parameters
    ----------
    path : Path
        Path to the LAMMPS dump file.

    Returns
    -------
    defaultdict
        The last snapshot from the LAMMPS dump file.
    """
    return deque(LAMMPSReader(path), maxlen=1).pop()



"""This module contains the `DamageDB` class."""

import warnings
from collections import deque
from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
from numpy import typing as npt
from numpy.lib.recfunctions import structured_to_unstructured as str2unstr
from scipy.spatial.transform import Rotation
from sklearn.decomposition import PCA

from irradiapy import dtypes, materials
from irradiapy.io.lammpsreader import LAMMPSReader


@dataclass
class DamageDB:
    """Class used to reconstruct the damage produced by a PKA from a database of MD debris.

    Attributes
    ----------
    dir_mddb : Path
        Directory of the MD debris database.
    compute_tdam : bool
        Whether to apply Lindhard's formula to the recoil energy. It should be `True` for
        MD simulations without electronic stopping.
    mat_pka : materials.Material
        PKA material.
    mat_target : materials.Material
        Target material.
    dpa_mode : materials.Material.DpaMode
        Mode for dpa calculation.
    tdam_mode : materials.Material.TdamMode
        Mode for PKA to damage energy calculation.
    seed : int, optional (default=0)
        Random seed for random number generator.
    """

    dir_mddb: Path
    compute_tdam: bool
    mat_pka: "materials.Material"
    mat_target: "materials.Material"
    dpa_mode: "materials.Material.DpaMode"
    tdam_mode: "materials.Material.TdamMode"
    seed: int = 0
    __rng: np.random.Generator = field(init=False)
    __calc_nd: callable = field(init=False)
    __files: dict[float, list[Path]] = field(init=False)
    __energies: npt.NDArray[np.float64] = field(init=False)
    __nenergies: int = field(init=False)

    def __post_init__(self) -> None:
        self.__rng = np.random.default_rng(self.seed)
        # Scan the database
        self.__files = {
            float(folder.name): list(folder.iterdir())
            for folder in self.dir_mddb.iterdir()
            if folder.is_dir()
        }
        self.__energies = np.array(sorted(self.__files.keys(), reverse=True))
        self.__nenergies = len(self.__energies)
        # PKA energy to damage energy conversion
        self.__compute_damage_energy = lambda x: self.mat_pka.epka_to_tdam(
            mat_pka=self.mat_target, epka=x, mode=self.tdam_mode
        )
        # Select the dpa model for residual energy
        self.__calc_nd = lambda x: self.mat_target.tdam_to_dpa(
            tdam=x, mode=self.dpa_mode
        )

    def __get_files(self, pka_e: float) -> tuple[dict[float, list[Path]], int]:
        """Get cascade files and number of residual FP for a given PKA energy.

        Parameters
        ----------
        pka_e : float
            PKA energy.

        Returns
        -------
        tuple[dict[float, list[Path]], int]
            Dictionary of selected paths and number of residual FP.
        """
        # Decompose the PKA energy into cascades and residual energy
        residual_energy = (
            self.__compute_damage_energy(pka_e) if self.compute_tdam else pka_e
        )
        cascade_counts = np.zeros(self.__nenergies, dtype=np.int64)
        for i, energy in enumerate(self.__energies):
            cascade_counts[i], residual_energy = divmod(residual_energy, energy)
        # Select the files for each energy
        if residual_energy > 0:
            residual_energy = self.__compute_damage_energy(residual_energy)
        debris_files = {
            energy: self.__rng.choice(self.__files[energy], cascade_counts[i])
            for i, energy in enumerate(self.__energies)
        }
        # Get the number of residual FP
        nfp = np.round(self.__calc_nd(residual_energy)).astype(np.int64)
        return debris_files, nfp

    def get_pka_debris(
        self,
        pka_e: float,
        pka_pos: npt.NDArray[np.float64],
        pka_dir: npt.NDArray[np.float64],
    ) -> dtypes.Defect:
        """Get PKA debris from its energy position, and direction.

        Parameters
        ----------
        pka_e : float
            PKA energy.
        pka_pos : npt.NDArray[np.float64]
            PKA position.
        pka_dir : npt.NDArray[np.float64]
            PKA direction.

        Returns
        -------

        dtypes.Defect
            Defects after the cascades.
        """
        files, nfp = self.__get_files(pka_e)
        # Get the maximum energy available in the database for the given PKA.
        # If no energy is available, return zero to place only FP.
        db_emax = next(
            (energy for energy in self.__energies if len(files[energy])), 0.0
        )
        # Possible to get cascades from the database
        if db_emax > 0.0:
            defects = self.__process_highest_energy_cascade(
                files, db_emax, pka_pos, pka_dir
            )
            parallelepiped = self.__get_parallelepiped(defects)
            defects = self.__place_other_debris(files, defects, parallelepiped)
            if nfp:
                defects = self.__place_fps_in_parallelepiped(
                    defects, nfp, parallelepiped
                )
            return defects
        # If no energy is available, generate FP only
        if nfp:
            defects = self.__place_fps_in_sphere(nfp, pka_pos, pka_dir)
            return defects
        defects = np.empty(0, dtype=dtypes.defect)
        return defects

    def __process_highest_energy_cascade(
        self,
        files: dict,
        db_emax: float,
        pka_pos: npt.NDArray[np.float64],
        pka_dir: npt.NDArray[np.float64],
    ) -> dtypes.Defect:
        """Process the highest energy cascade.

        Parameters
        ----------
        files : dict
            Dictionary of files for each energy.
        db_emax : float
            Energy of the highest energy cascade.
        pka_pos : npt.NDArray[np.float64]
            PKA position.
        pka_dir : npt.NDArray[np.float64]
            PKA direction.

        Returns
        -------
        dtypes.Defect
            Defects after the highest energy cascade.
        """
        file = files[db_emax][0]
        files[db_emax] = np.delete(files[db_emax], 0)
        defects = deque(LAMMPSReader(file), maxlen=1).pop()["atoms"]
        xaxis = np.array([1.0, 0.0, 0.0])
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            transform = Rotation.align_vectors([pka_dir], [xaxis])[0]

        pos = str2unstr(defects[["x", "y", "z"]], dtype=np.float64, copy=False)
        pos = transform.apply(pos) + pka_pos
        defects["x"] = pos[:, 0]
        defects["y"] = pos[:, 1]
        defects["z"] = pos[:, 2]

        return defects

    def __place_other_debris(
        self,
        files: dict,
        defects: dtypes.Defect,
        parallelepiped: tuple[PCA, npt.NDArray, npt.NDArray],
    ) -> dtypes.Defect:
        """Place other debris in the parallelepiped.

        Parameters
        ----------
        files : dict
            Dictionary of files for each energy.
        defects : dtypes.Defect
            Defects after the highest energy cascade.
        parallelepiped : tuple[PCA, npt.NDArray, npt.NDArray]
            Parallelepiped definition.

        Returns
        -------
        dtypes.Defect
            Defects after placing the other debris.
        """
        for energy in self.__energies:
            for file0 in files[energy]:
                defects_ = deque(LAMMPSReader(file0), maxlen=1).pop()["atoms"]

                transform = Rotation.random(rng=self.__rng)
                pos = str2unstr(defects_[["x", "y", "z"]], dtype=np.float64, copy=False)
                pos0 = self.__get_parallelepiped_points(*parallelepiped, 1)
                pos = transform.apply(pos) + pos0
                defects_["x"] = pos[:, 0]
                defects_["y"] = pos[:, 1]
                defects_["z"] = pos[:, 2]

                defects = np.concatenate((defects, defects_))
        return defects

    def __place_fps_in_parallelepiped(
        self,
        defects: dtypes.Defect,
        nfp: int,
        parallelepiped: tuple[PCA, npt.NDArray, npt.NDArray],
    ) -> dtypes.Defect:
        """Place FPs anywhere in the parallelepiped.

        Parameters
        ----------
        defects : dtypes.Defect
            Defects after placing the other debris.
        nfp : int
            Number of FPs.
        parallelepiped : tuple[PCA, npt.NDArray, npt.NDArray]
            Parallelepiped definition.

        Returns
        -------
        dtypes.Defect
            Defects after placing the FPs.
        """
        defects_ = np.zeros(2 * nfp, dtype=dtypes.defect)
        defects_["type"][:nfp] = self.mat_target.atomic_number
        defects_["x"][:nfp] = self.mat_target.dist_fp / 2.0
        pos = str2unstr(defects_[["x", "y", "z"]], dtype=np.float64, copy=False)
        pos[:nfp] = Rotation.random(nfp, rng=self.__rng).apply(pos[:nfp])
        pos[nfp:] = -pos[:nfp]
        pos0 = self.__get_parallelepiped_points(*parallelepiped, nfp)
        pos[:nfp] += pos0
        pos[nfp:] += pos0
        defects_["x"] = pos[:, 0]
        defects_["y"] = pos[:, 1]
        defects_["z"] = pos[:, 2]

        return np.concatenate((defects, defects_))

    def __place_fps_in_sphere(
        self,
        nfp: int,
        pka_pos: npt.NDArray[np.float64],
        pka_dir: npt.NDArray[np.float64],
    ) -> dtypes.Defect:
        """Generate FPs in a sphere.

        Parameters
        ----------
        nfp : int
            Number of FPs.
        pka_pos : npt.NDArray[np.float64]
            PKA position.
        pka_dir : npt.NDArray[np.float64]
            PKA direction.

        Returns
        -------
        dtypes.Defect
            Defects after generating.
        """
        defects_ = np.zeros(2 * nfp, dtype=dtypes.defect)
        defects_["type"][:nfp] = self.mat_target.atomic_number
        defects_["x"][:nfp] = self.mat_target.dist_fp / 2
        pos = str2unstr(defects_[["x", "y", "z"]], dtype=np.float64, copy=False)
        pos[:nfp] = Rotation.random(nfp, rng=self.__rng).apply(pos[:nfp])
        pos[nfp:] = -pos[:nfp]

        random = self.__rng.random((nfp, 3))
        theta = np.arccos(2.0 * random[:, 0] - 1.0)
        phi = 2.0 * np.pi * random[:, 1]
        radius = nfp * self.mat_target.dist_fp / 2.0
        r = radius * np.cbrt(random[:, 2])
        points = np.empty((nfp, 3))
        points[:, 0] = r * np.sin(theta) * np.cos(phi)
        points[:, 1] = r * np.sin(theta) * np.sin(phi)
        points[:, 2] = r * np.cos(theta)

        pos[:nfp] += points
        pos[nfp:] += points
        pos += pka_pos + pka_dir * radius
        defects_["x"] = pos[:, 0]
        defects_["y"] = pos[:, 1]
        defects_["z"] = pos[:, 2]

        return defects_

    def __get_parallelepiped(self, atoms: dtypes.Atom) -> tuple:
        """
        Define a parallelepiped from the atomic positions using PCA.

        Parameters
        ----------
        atoms : dtypes.Atom
            Atomic positions.

        Returns
        -------
        tuple
            PCA object, minimum PCA coordinates, maximum PCA coordinates.
        """
        pos = str2unstr(atoms[["x", "y", "z"]], dtype=np.float64, copy=False)
        pca = PCA(n_components=3)
        pca.fit(pos)
        atoms_pca = pca.transform(pos)
        min_pca = np.min(atoms_pca, axis=0)
        max_pca = np.max(atoms_pca, axis=0)
        return pca, min_pca, max_pca

    def __get_parallelepiped_points(
        self,
        pca: PCA,
        min_pca: npt.NDArray[np.float64],
        max_pca: npt.NDArray[np.float64],
        npoints: int,
    ) -> npt.NDArray[np.float64]:
        """
        Generate random points within a parallelepiped.

        Parameters
        ----------
        pca : PCA
            PCA object.
        min_pca : npt.NDArray[np.float64]
            Minimum PCA coordinates.
        max_pca : npt.NDArray[np.float64]
            Maximum PCA coordinates.
        npoints : int
            Number of points to generate.

        Returns
        -------
        npt.NDArray[np.float64]
            Random points within the parallelepiped.
        """
        random_points_pca = self.__rng.uniform(min_pca, max_pca, size=(npoints, 3))
        return pca.inverse_transform(random_points_pca)



"""Module for configuration variables."""

from pathlib import Path

import matplotlib.pyplot as plt

# region General

#: str: Format for integers in output files.
INT_FORMAT = "%d"
#: str: Format for floats in output files.
FLOAT_FORMAT = "%g"
#: str: Encoding for text files.
ENCODING = "utf-8"
#: list[str]: List of atom fields to exclude from output in LAMMPS files.
EXCLUDED_ITEMS = ["xs", "ys", "zs"]


def use_style(latex: bool = False) -> None:
    """Set the style for matplotlib plots.

    It uses the colour universal design (CUD) palette for colour-blind friendly plots.

    Parameters
    ----------
    latex : bool, optional (default=False)
        If True, use LaTeX for text rendering in plots (slower). I might require other software to
        be installed on your system.
    """
    if latex:
        plt.style.use("irradiapy.styles.latex")
    else:
        plt.style.use("irradiapy.styles.nolatex")


# region SRIM

#: pathlib.Path: TRIM.exe directory (parent folder)
DIR_SRIM = Path("./SRIM-2013")



"""This module contains the `LAMMPSWriterMPI` class."""

# pylint: disable=no-name-in-module, broad-except

from dataclasses import dataclass, field
from pathlib import Path
from types import TracebackType
from typing import Any, TextIO

import numpy.typing as npt
from mpi4py import MPI

from irradiapy import config
from irradiapy.utils.mpi import (
    MPIExceptionHandlerMixin,
    MPITagAllocator,
    mpi_safe_method,
)


@dataclass
class LAMMPSWriterMPI(MPIExceptionHandlerMixin):
    """A class to write data like a LAMMPS dump file in parallel using MPI.

    Attributes
    ----------
    file_path : Path
        The path to the LAMMPS dump file.
    mode : str, optional (default="w")
        The file open mode.
    excluded_items : list[str], optional (default=irradiapy.config.EXCLUDED_ITEMS)
        Atom fields to exclude from output.
    encoding : str, optional (default=irradiapy.config.ENCODING)
        The file encoding.
    int_format : str, optional (default=irradiapy.config.INT_FORMAT)
        The format for integers.
    float_format : str, optional (default=irradiapy.config.FLOAT_FORMAT)
        The format for floats.
    comm : MPI.Comm, optional (default=mpi4py.MPI.COMM_WORLD)
        The MPI communicator.
    """

    file_path: Path
    mode: str = "w"
    excluded_items: list[str] = field(default_factory=lambda: config.EXCLUDED_ITEMS)
    encoding: str = field(default_factory=lambda: config.ENCODING)
    int_format: str = field(default_factory=lambda: config.INT_FORMAT)
    float_format: str = field(default_factory=lambda: config.FLOAT_FORMAT)
    file: TextIO = field(default=None, init=False)
    comm: MPI.Comm = field(default_factory=lambda: MPI.COMM_WORLD)
    __rank: int = field(init=False)
    __commsize: int = field(init=False)
    __comm_tag: int = field(default_factory=MPITagAllocator.get_tag, init=False)

    def __post_init__(self) -> None:
        """Opens the file associated with this writer."""
        self.__rank = self.comm.Get_rank()
        self.__commsize = self.comm.Get_size()
        self.file = None
        if self.__rank == 0:
            try:
                self.file = open(self.file_path, self.mode, encoding=self.encoding)
            except Exception:
                self._handle_exception()

    def __enter__(self) -> "LAMMPSWriterMPI":
        return self

    def __del__(self) -> None:
        if self.__rank == 0 and self.file is not None:
            self.file.close()

    def __exit__(
        self,
        exc_type: None | type[BaseException] = None,
        exc_value: None | BaseException = None,
        exc_traceback: None | TracebackType = None,
    ) -> bool:
        if self.__rank == 0 and self.file is not None:
            self.file.close()
        return False

    def __atoms_rank_to_string(
        self, atoms_rank: npt.NDArray, field_names: list[str], formatters: list[str]
    ) -> str:
        """Converts the atoms_rank array to a formatted string.

        Parameters
        ----------
        atoms_rank : npt.NDArray
            The atoms_rank array to be converted.
        field_names : list[str]
            The names of the fields in the structured array.
        formatters : list[str]
            The format strings for each field.

        Returns
        -------
        str
            A formatted string representation of the atoms_rank array.
        """
        lines_chunk = "\n".join(
            " ".join(
                fmt % atom[field_name]
                for fmt, field_name in zip(formatters, field_names)
            )
            for atom in atoms_rank
        )
        return lines_chunk

    @mpi_safe_method
    def close(self) -> None:
        """Closes the file associated with this writer."""
        if self.__rank == 0 and not self.file.closed:
            self.file.close()

    @mpi_safe_method
    def write(self, data: dict[str, Any]) -> None:
        """Writes the data to the file.

        Note
        ----
        Assumes orthogonal simulation box.

        Parameters
        ----------
        data : dict[str, Any]
            A dictionary containing the data to be written. The keys should
            include 'timestep', 'boundary', 'xlo', 'xhi', 'ylo', 'yhi', 'zlo', 'zhi',
            and any other fields to be written as atom properties.
        """
        atoms = data["atoms"]
        field_names = [f for f in atoms.dtype.names if f not in self.excluded_items]

        formatters = []
        for field_name in field_names:
            dtype = atoms.dtype[field_name]
            if dtype.kind == "i":
                formatters.append(self.int_format)
            elif dtype.kind == "f":
                formatters.append(self.float_format)
            else:
                formatters.append("%s")

        if self.__rank == 0:
            if "time" in data:
                self.file.write(f"ITEM: TIME\n{data['time']}\n")
            self.file.write(f"ITEM: TIMESTEP\n{data['timestep']}\n")
            self.file.write(f"ITEM: NUMBER OF ATOMS\n{data['natoms']}\n")
            self.file.write(f"ITEM: BOX BOUNDS {' '.join(data['boundary'])}\n")
            self.file.write(
                f"{self.float_format % data['xlo']} {self.float_format % data['xhi']}\n"
            )
            self.file.write(
                f"{self.float_format % data['ylo']} {self.float_format % data['yhi']}\n"
            )
            self.file.write(
                f"{self.float_format % data['zlo']} {self.float_format % data['zhi']}\n"
            )
            self.file.write(f"ITEM: ATOMS {' '.join(field_names)}\n")

        self.comm.Barrier()
        lines_chunk = self.__atoms_rank_to_string(
            data["atoms"], field_names, formatters
        )
        if self.__rank == 0:
            self.file.write(lines_chunk)
            if lines_chunk:
                self.file.write("\n")
            for sender_rank in range(1, self.__commsize):
                self.comm.send(None, dest=sender_rank, tag=self.__comm_tag + 1)
                msg = self.comm.recv(source=sender_rank, tag=self.__comm_tag)
                self.file.write(msg)
                if msg:
                    self.file.write("\n")
        else:
            self.comm.recv(source=0, tag=self.__comm_tag + 1)
            self.comm.send(lines_chunk, dest=0, tag=self.__comm_tag)
        self.comm.Barrier()



"""Defects analysis module."""

from collections import defaultdict
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.gridspec import GridSpec
from numpy import typing as npt

from irradiapy import dtypes, materials, utils
from irradiapy.analysis.defectsidentifier import DefectsIdentifier
from irradiapy.io import BZIP2LAMMPSReader, LAMMPSReader, LAMMPSWriter


def identify_defects(
    lattice: str,
    a0: float,
    data_atoms: defaultdict,
    a1: None | float = None,
    pos_pka: None | npt.NDArray[np.float64] = None,
    theta_pka: None | float = None,
    phi_pka: None | float = None,
    transform: None | bool = False,
    debug: bool = False,
) -> dtypes.Defect:
    """Identify defects in a given atomic structure.

    Parameters
    ----------
    lattice : str
        Lattice type. Currently only "bcc" is supported.
    a0 : float
        Lattice parameter.
    data_atoms : defaultdict
        Dictionary containing simulation data as given by the LAMMPSReader and similar readers.
        Must include keys: 'atoms', 'natoms', 'boundary', 'xlo', 'xhi', 'ylo', 'yhi', 'zlo',
        'zhi', 'timestep'.
    a1 : float, optional (default=None)
        Final lattice parameter. If provided, defect positions are rescaled to this value
        (independently of the `transform` value).
    pos_pka : npt.NDArray[np.float64], optional (default=None)
        Position vector of the PKA. If provided with theta_pka and phi_pka, defects are
        recentered and aligned.
    theta_pka : float, optional (default=None)
        Polar angle (in radians) for the PKA direction.
    phi_pka : float, optional (default=None)
        Azimuthal angle (in radians) for the PKA direction.
    transform : bool, optional (default=False)
        If True, defects are recentered and aligned with the PKA direction (if provided). If
        True but no PKA parameters are provided, defects are recentered based on their
        average position. Note that the box boundaries are not modified for visualization
        purposes, only the atomic positions are transformed.
    debug : bool, optional (default=False)
        If `True`, enables debug mode for additional output.

    Returns
    -------
    dtypes.Defect
        An array of identified defects in the structure.
    """
    defects_finder = DefectsIdentifier(lattice=lattice, a0=a0, debug=debug)
    defects = defects_finder.identify(
        data_atoms=data_atoms,
        a1=a1,
        pos_pka=pos_pka,
        theta_pka=theta_pka,
        phi_pka=phi_pka,
        transform=transform,
    )
    return defects


def identify_lammps_dump(
    lattice: str,
    a0: float,
    path_dump: Path,
    path_dump_defects: Path,
    a1: None | float = None,
    pos_pka: None | npt.NDArray[np.float64] = None,
    theta_pka: None | float = None,
    phi_pka: None | float = None,
    transform: None | bool = False,
    overwrite: bool = False,
    debug: bool = False,
) -> None:
    """Identify defects in a LAMMPS dump file.

    Parameters
    ----------
    lattice : str
        Lattice type. Currently only "bcc" is supported.
    a0 : float
        Lattice parameter.
    data_atoms : defaultdict
        Dictionary containing simulation data as given by the LAMMPSReader and similar readers.
    data_atoms : defaultdict
        Dictionary containing simulation data as given by the LAMMPSReader and similar readers.
        Must include keys: 'atoms', 'natoms', 'boundary', 'xlo', 'xhi', 'ylo', 'yhi', 'zlo',
        'zhi', 'timestep'.
    path_dump : Path
        Path to the LAMMPS dump file to read. Can be compressed with `.bz2` or not.
    path_dump_defects : Path
        Path to the output file where identified defects will be written (in text format).
    a1 : float, optional
        Final lattice parameter. If provided, defect positions are rescaled to this value
        (independently of the `transform` value).
    pos_pka : npt.NDArray[np.float64], optional
        Position vector of the PKA. If provided with theta_pka and phi_pka, defects are
        recentered and aligned.
    theta_pka : float, optional
        Polar angle (in radians) for the PKA direction.
    phi_pka : float, optional
        Azimuthal angle (in radians) for the PKA direction.
    transform : bool, optional
        If True, defects are recentered and aligned with the PKA direction (if provided). If
        True but no PKA parameters are provided, defects are recentered based on their
        average position. Note that the box boundaries are not modified for visualization
        purposes, only the atomic positions are transformed.
    overwrite : bool, optional (default=False)
        If True, allows overwriting the output file if it already exists.
    debug : bool, optional (default=False)
        If `True`, enables debug mode for additional output.
    """
    if path_dump_defects.exists():
        if overwrite:
            path_dump_defects.unlink()
        else:
            raise FileExistsError(f"Defects file {path_dump_defects} already exists.")
    if debug:
        print(f"Identifying defects in {path_dump}")
    if path_dump.suffix == ".bz2":
        reader = BZIP2LAMMPSReader(path_dump)
    else:
        reader = LAMMPSReader(path_dump)
    writer = LAMMPSWriter(path_dump_defects, mode="a")
    defects_finder = DefectsIdentifier(lattice=lattice, a0=a0, debug=debug)
    for data_atoms in reader:
        if debug:
            print(f"Timestep {data_atoms['timestep']}")
        defects = defects_finder.identify(
            data_atoms,
            a1=a1,
            pos_pka=pos_pka,
            theta_pka=theta_pka,
            phi_pka=phi_pka,
            transform=transform,
        )
        writer.write(defects)


def plot_mddb_nd(
    target_dir: Path,
    mat_pka: materials.Material,
    mat_target: materials.Material,
    path_plot: Path,
    dpi: int = 300,
) -> None:
    """Plot the number of defects (vacancies) as a function of the PKA energy from a molecular
    dynamics database.

    Note
    ----
    The number of defects is estimated from the vacancy counts.

    Parameters
    ----------
    target_dir : Path
        Directory containing the MD database.
    mat_pka : materials.Material
        Material of the PKA.
    mat_target : materials.Material
        Target material.
    path_plot : Path, optional
        Path to save the plot. If `None`, the plot is shown but not saved.
    dpi : int, optional (default=300)
        Dots per inch.
    """

    # Extract data from xyz files
    nd_all = defaultdict(defaultdict)
    nfiles = defaultdict(int)
    for energy_dir in target_dir.glob("*"):
        if not energy_dir.is_dir():
            continue
        energy = int(energy_dir.name)
        nd_all[energy] = {
            "nd": [],
            "mean": 0,
            "std": 0,
            "ste": 0,
        }
        for path_defects in energy_dir.glob("*.xyz"):
            nfiles[energy] += 1
            data_defects = utils.io.get_last_lammps_dump(path_defects)

            cond = data_defects["atoms"]["type"] == 0
            vacs = data_defects["atoms"][cond]
            nd_all[energy]["nd"].append(len(vacs))
        nd_all[energy]["mean"] = np.mean(nd_all[energy]["nd"])
        nd_all[energy]["std"] = np.std(nd_all[energy]["nd"])
        nd_all[energy]["ste"] = nd_all[energy]["std"] / np.sqrt(nfiles[energy])

    # Sort energies
    epkas = np.array(
        [int(path.name) for path in target_dir.glob("*") if path.is_dir()],
        dtype=np.int64,
    )
    epkas.sort()

    # Calculate NRT- and fer-arc-dpa
    range_epkas = np.linspace(epkas[0], epkas[-1], 1000)

    range_tdams = mat_target.epka_to_tdam(mat_pka, range_epkas)
    nrt_dpa = mat_target.calc_nrt_dpa(range_tdams)
    fer_arc_dpa = mat_target.calc_fer_arc_dpa(range_tdams)

    fig = plt.figure(figsize=(10, 6))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel("Energy (keV)")
    ax.set_ylabel("$N_d$")

    axins = ax.inset_axes([0.05, 0.55, 0.4, 0.4])
    # Set the xlim and ylim for the inset in keV
    x1, x2 = 0, 21
    ydata = np.array(
        [nd_all[epka]["mean"] for epka in epkas],
        dtype=np.float64,
    )
    # Find the max y in the zoomed x region
    mask = epkas / 1e3 <= x2
    if np.any(mask):
        ymax = max(ydata[mask]) * 1.1
    else:
        ymax = max(ydata) * 1.1
    axins.set_xlim(x1, x2)
    axins.set_ylim(0.0, ymax)
    _, connectors = ax.indicate_inset_zoom(axins)
    for connector in connectors:  # otherwise the connectors are not visible
        connector.set_visible(True)

    ax.errorbar(
        epkas / 1e3,
        [nd_all[energy]["mean"] for energy in epkas],
        yerr=[nd_all[energy]["ste"] for energy in epkas],
        marker="o",
        linestyle="none",
        label="MD DB",
    )

    ax.plot(
        range_epkas / 1e3,
        nrt_dpa,
        label="NRT-dpa",
    )
    ax.plot(
        range_epkas / 1e3,
        fer_arc_dpa,
        label="fer-arc-dpa",
    )

    axins.errorbar(
        epkas / 1e3,
        [nd_all[energy]["mean"] for energy in epkas],
        yerr=[nd_all[energy]["ste"] for energy in epkas],
        marker="o",
        linestyle="none",
    )
    axins.plot(
        range_epkas / 1e3,
        nrt_dpa,
    )
    axins.plot(
        range_epkas / 1e3,
        fer_arc_dpa,
    )

    ax.legend(loc="lower right")
    fig.tight_layout()
    if path_plot is not None:
        fig.savefig(path_plot, dpi=dpi)
    else:
        plt.show()
    plt.close(fig)



"""This module provides a class to find and analyze defects in crystalline structures."""

from collections import defaultdict
from dataclasses import dataclass, field

import numpy as np
import numpy.typing as npt
from numpy.lib.recfunctions import structured_to_unstructured as str2unstr
from scipy.spatial.transform import Rotation as R

from irradiapy import dtypes


@dataclass
class DefectsIdentifier:
    """Class to identify  defects in crystalline structures.

    This class provides methods to identify point defects (vacancies and interstitials) in a
    body-centered cubic (bcc) lattice based on atomic positions from simulation data. It supports
    optional rescaling and recentering of defect positions, and can align the system with a
    specified primary knock-on atom (PKA) direction.

    Parameters
    ----------
    lattice : str
        Lattice type. Currently only "bcc" is supported.
    a0 : float
        Lattice parameter, in angstroms.
    debug : bool, optional (default=False)
        If True, enables debug mode for additional output.
    """

    lattice: str
    a0: float
    debug: bool = False
    __perx: bool = field(init=False)
    __pery: bool = field(init=False)
    __perz: bool = field(init=False)
    __nxhi: int = field(init=False)
    __nyhi: int = field(init=False)
    __nzhi: int = field(init=False)
    __nxlo: int = field(init=False)
    __nylo: int = field(init=False)
    __nzlo: int = field(init=False)
    __nx: int = field(init=False)
    __ny: int = field(init=False)
    __nz: int = field(init=False)
    __sub_count: int = 2  # Number of atoms per primitive unit cell (bcc).

    def __post_init__(self) -> None:
        if self.lattice == "bcc":
            self.__sub_count = 2
            self.__pos_ucell = np.array(
                [
                    [0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0],
                    [0.0, 1.0, 0.0],
                    [0.0, 0.0, 1.0],
                    [1.0, 0.0, 1.0],
                    [1.0, 1.0, 1.0],
                    [0.0, 1.0, 1.0],
                    [0.5, 0.5, 0.5],
                ],
                dtype=np.float64,
            )
            self.__idx_ucell = np.array(
                [
                    [0, 0, 0, 0],
                    [1, 0, 0, 0],
                    [1, 1, 0, 0],
                    [0, 1, 0, 0],
                    [0, 0, 1, 0],
                    [1, 0, 1, 0],
                    [1, 1, 1, 0],
                    [0, 1, 1, 0],
                    [0, 0, 0, 1],
                ],
                dtype=np.int64,
            )
            # Scale positions by a0
            self.__pos_ucell *= self.a0
        else:
            raise ValueError("Only bcc lattice is supported.")

    def __rescale_translate_rotate(
        self,
        atoms: dtypes.Atom,
        a1: float | np.number,
        pos_pka: npt.NDArray[np.float64],
        theta_pka: float,
        phi_pka: float,
    ) -> None:
        """Rescales, translates, and rotates the positions of atoms.

        Parameters
        ----------
        atoms : dtypes.Atom
            Structured array of atomic positions (fields: x, y, z).
        a1 : float
            Final lattice parameter to rescale positions.
        pos_pka : npt.NDArray[np.float64]
            Position vector of the PKA for translation.
        theta_pka : float
            Polar angle (in radians) for the PKA direction (for rotation).
        phi_pka : float
            Azimuthal angle (in radians) for the PKA direction (for rotation).
        """
        # Translate
        atoms["x"] -= pos_pka[0]
        atoms["y"] -= pos_pka[1]
        atoms["z"] -= pos_pka[2]

        # Rotation matrix, align with PKA initial direction
        xaxis = np.array([1.0, 0.0, 0.0])
        pka_dir = np.array(
            [
                np.sin(theta_pka) * np.cos(phi_pka),
                np.sin(theta_pka) * np.sin(phi_pka),
                np.cos(theta_pka),
            ]
        )
        transform = R.align_vectors([xaxis], [pka_dir])[0].as_matrix()

        # Scaling matrix
        scaling = a1 / self.a0
        scaling_matrix = (
            np.diag([scaling] * 3) if isinstance(scaling, float) else np.diag(scaling)
        )

        # Combined scale + rotation
        transform = transform @ scaling_matrix

        # Apply transformations
        pos = str2unstr(atoms[["x", "y", "z"]])
        pos = transform.apply(pos)
        atoms["x"] = pos[:, 0]
        atoms["y"] = pos[:, 1]
        atoms["z"] = pos[:, 2]

    def __apply_boundary_conditions(self, idx_atoms: npt.NDArray[np.int64]) -> None:
        """Applies periodic boundary conditions to atomic index coordinates.

        Only one unit cell around the simulation box is considered for periodic boundary conditions.

        Parameters
        ----------
        idx_atoms : npt.NDArray[np.int64]
            Array of atomic positions with index coordinates (shape: [N, 4]).
        """
        if self.__perx:
            idx_atoms[:, 0][idx_atoms[:, 0] == self.__nxhi] = self.__nxlo
            idx_atoms[:, 0][idx_atoms[:, 0] == self.__nxlo - 1] = self.__nxhi - 1
        if self.__pery:
            idx_atoms[:, 1][idx_atoms[:, 1] == self.__nyhi] = self.__nylo
            idx_atoms[:, 1][idx_atoms[:, 1] == self.__nylo - 1] = self.__nyhi - 1
        if self.__perz:
            idx_atoms[:, 2][idx_atoms[:, 2] == self.__nzhi] = self.__nzlo
            idx_atoms[:, 2][idx_atoms[:, 2] == self.__nzlo - 1] = self.__nzhi - 1

    def __site_id_to_indices(self, i: int) -> tuple[int, int, int, int]:
        """Decodes a site ID into its corresponding lattice indices.

        Parameters
        ----------
        i : int
            Site ID to decode.

        Returns
        -------
        tuple of int
            (ix, iy, iz, ia) indices corresponding to the site ID.
        """
        ia = i % self.__sub_count
        tmp = i // self.__sub_count
        iz = tmp % self.__nz
        tmp //= self.__nz
        iy = tmp % self.__ny
        ix = tmp // self.__ny
        ix += self.__nxlo
        iy += self.__nylo
        iz += self.__nzlo
        return ix, iy, iz, ia

    def __site_id_to_cartesian(self, i: int) -> npt.NDArray[np.float64]:
        """Converts a site ID to its corresponding cartesian coordinates.

        Parameters
        ----------
        i : int
            Site ID to decode.

        Returns
        -------
        npt.NDArray[np.float64]
            Cartesian coordinates (x, y, z) corresponding to the site ID.
        """
        ix, iy, iz, ia = self.__site_id_to_indices(i)
        x = (ix + 0.5 * ia) * self.a0
        y = (iy + 0.5 * ia) * self.a0
        z = (iz + 0.5 * ia) * self.a0
        return np.array([x, y, z], dtype=np.float64)

    def __defect_identification(
        self,
        data_atoms: defaultdict,
        idx_atoms: npt.NDArray[np.int64],
    ) -> dtypes.Defect:
        """Identifies defects based on the assignments.

        Parameters
        ----------
        data_atoms : defaultdict
            Dictionary containing simulation data as given by the LAMMPSReader and similar readers.
        idx_atoms : npt.NDArray[np.int64]
            Array of atomic index coordinates (shape: [N, 4]).

        Returns
        -------
        dtypes.Defect
            Array of defects (structured array with fields: type, x, y, z).
        """
        # Build unique site IDs from idx_atoms
        ix = idx_atoms[:, 0] - self.__nxlo
        iy = idx_atoms[:, 1] - self.__nylo
        iz = idx_atoms[:, 2] - self.__nzlo
        s = idx_atoms[:, 3]
        nsites = self.__nx * self.__ny * self.__nz * self.__sub_count
        id_sites = ((ix * self.__ny + iy) * self.__nz + iz) * self.__sub_count + s

        # Sort and group
        sort_idx = np.argsort(id_sites)
        site_sorted = id_sites[sort_idx]
        split_points = np.cumsum(np.bincount(site_sorted, minlength=nsites))[:-1]
        grouped = np.split(sort_idx, split_points)

        defects = np.empty(0, dtype=dtypes.defect)
        for i, grp in enumerate(grouped):
            if len(grp) == 0:
                # Vacancy
                vac = np.array(
                    [(0, *self.__site_id_to_cartesian(i))], dtype=dtypes.defect
                )
                defects = np.concatenate((defects, vac))
            elif len(grp) > 1:
                # Interstitials
                xyz = self.__site_id_to_cartesian(i)
                coords = str2unstr(data_atoms["atoms"][["x", "y", "z"]][grp])
                dist2 = np.sum(np.square(coords - xyz), axis=1)
                keep_idx = np.argmin(dist2)
                inters_idx = np.ones(len(grp), dtype=bool)
                inters_idx[keep_idx] = False
                if np.any(inters_idx):
                    count = np.count_nonzero(inters_idx)
                    inters = np.zeros(count, dtype=dtypes.defect)
                    inters["type"] = data_atoms["atoms"]["element"][grp][inters_idx]
                    inters["x"] = coords[inters_idx, 0]
                    inters["y"] = coords[inters_idx, 1]
                    inters["z"] = coords[inters_idx, 2]
                    defects = np.concatenate((defects, inters))
        return defects

    def identify(
        self,
        data_atoms: defaultdict,
        a1: None | float = None,
        pos_pka: None | npt.NDArray[np.float64] = None,
        theta_pka: None | float = None,
        phi_pka: None | float = None,
        transform: None | bool = False,
    ) -> defaultdict:
        """Identify defects in the crystalline structure based on atomic positions.

        Parameters
        ----------
        data_atoms : defaultdict
            Dictionary containing simulation data as given by the LAMMPSReader and similar readers.
            Must include keys: 'atoms', 'natoms', 'boundary', 'xlo', 'xhi', 'ylo', 'yhi', 'zlo',
            'zhi', 'timestep'.
        a1 : float, optional
            Final lattice parameter. If provided, defect positions are rescaled to this value
            (independently of the `transform` value).
        pos_pka : npt.NDArray[np.float64], optional
            Position vector of the PKA. If provided with theta_pka and phi_pka, defects are
            recentered and aligned.
        theta_pka : float, optional
            Polar angle (in radians) for the PKA direction.
        phi_pka : float, optional
            Azimuthal angle (in radians) for the PKA direction.
        transform : bool, optional
            If True, defects are recentered and aligned with the PKA direction (if provided). If
            True but no PKA parameters are provided, defects are recentered based on their
            average position. Note that the box boundaries are not modified for visualization
            purposes, only the atomic positions are transformed.

        Returns
        -------
        data_defects : defaultdict
            Dictionary containing the defects found in the structure. Keys are the same as in
            `data_atoms`, but the 'atoms' key contains only defects and
            'natoms' reflects the number of defects found.
        """
        self.__nxlo = round(data_atoms["xlo"] / self.a0)
        self.__nxhi = round(data_atoms["xhi"] / self.a0)
        self.__nylo = round(data_atoms["ylo"] / self.a0)
        self.__nyhi = round(data_atoms["yhi"] / self.a0)
        self.__nzlo = round(data_atoms["zlo"] / self.a0)
        self.__nzhi = round(data_atoms["zhi"] / self.a0)
        self.__nx = self.__nxhi - self.__nxlo
        self.__ny = self.__nyhi - self.__nylo
        self.__nz = self.__nzhi - self.__nzlo
        self.__perx = data_atoms["boundary"][0] == "pp"
        self.__pery = data_atoms["boundary"][1] == "pp"
        self.__perz = data_atoms["boundary"][2] == "pp"

        idx_atoms = np.zeros((data_atoms["natoms"], 4), dtype=np.int64)
        mod_atoms = np.zeros((data_atoms["natoms"], 3), dtype=np.float64)
        idx_atoms[:, 0], mod_atoms[:, 0] = np.divmod(data_atoms["atoms"]["x"], self.a0)
        idx_atoms[:, 1], mod_atoms[:, 1] = np.divmod(data_atoms["atoms"]["y"], self.a0)
        idx_atoms[:, 2], mod_atoms[:, 2] = np.divmod(data_atoms["atoms"]["z"], self.a0)
        # idx_atoms[:, 3] is initially zero

        # Compute closest unit-cell positions
        cell_xyz = self.__pos_ucell
        dist = np.sum((mod_atoms[:, None, :] - cell_xyz[None, :, :]) ** 2, axis=2)
        closest_cell_indices = np.argmin(dist, axis=1)
        idx_atoms[:, 0] += self.__idx_ucell[closest_cell_indices, 0]
        idx_atoms[:, 1] += self.__idx_ucell[closest_cell_indices, 1]
        idx_atoms[:, 2] += self.__idx_ucell[closest_cell_indices, 2]
        idx_atoms[:, 3] += self.__idx_ucell[closest_cell_indices, 3]

        # Apply boundaries on idx_atoms
        self.__apply_boundary_conditions(idx_atoms)

        # Identify defects
        defects = self.__defect_identification(data_atoms, idx_atoms)

        # Recenter / scaling if requested
        if len(defects):
            if transform:
                if a1 is None:
                    a1 = self.a0
                if (
                    pos_pka is not None
                    and theta_pka is not None
                    and phi_pka is not None
                ):
                    self.__rescale_translate_rotate(
                        defects, a1, pos_pka, theta_pka, phi_pka
                    )
                else:
                    defects[["x", "y", "z"]] -= np.mean(
                        defects[["x", "y", "z"]], axis=1
                    )
                    defects[["x", "y", "z"]] *= a1
            elif a1 is not None:
                defects[["x", "y", "z"]] *= a1

        data_defects = defaultdict(None)
        data_defects["time"] = data_atoms["time"]
        data_defects["timestep"] = data_atoms["timestep"]
        data_defects["natoms"] = len(defects)
        data_defects["boundary"] = data_atoms["boundary"]
        data_defects["xlo"] = data_atoms["xlo"]
        data_defects["xhi"] = data_atoms["xhi"]
        data_defects["ylo"] = data_atoms["ylo"]
        data_defects["yhi"] = data_atoms["yhi"]
        data_defects["zlo"] = data_atoms["zlo"]
        data_defects["zhi"] = data_atoms["zhi"]
        data_defects["atoms"] = defects

        nvacs = np.count_nonzero(defects["type"] == 0)
        nsias = len(defects) - nvacs
        if self.debug:
            print(f"Number of interstitials: {nsias}")
            print(f"Number of vacancies: {nvacs}")
        return data_defects



"""This module contains the `LAMMPSReaderMPI` class."""

# pylint: disable=no-name-in-module, broad-except

from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from types import TracebackType
from typing import Any, Dict, Generator, TextIO, Tuple, Type

import numpy as np
from mpi4py import MPI
from numpy import typing as npt

from irradiapy.utils.mpi import (
    MPIExceptionHandlerMixin,
    MPITagAllocator,
    mpi_safe_method,
    mpi_subdomains_decomposition,
)


@dataclass
class LAMMPSReaderMPI(MPIExceptionHandlerMixin):
    """A class to read data from a LAMMPS dump  file in parallel using MPI.

    Note
    ----
    Assumed orthogonal simulation box.

    Attributes
    ----------
    file_path : Path
        The path to the LAMMPS dump file.
    encoding : str, optional (default="utf-8")
        The file encoding.
    comm : MPI.Comm, optional (default=mpi4py.MPI.COMM_WORLD)
        The MPI communicator.
    """

    file_path: Path
    encoding: str = "utf-8"
    file: TextIO = field(default=None, init=False)
    comm: MPI.Comm = field(default_factory=lambda: MPI.COMM_WORLD)
    __rank: int = field(init=False)
    __commsize: int = field(init=False)
    __comm_tag: int = field(default_factory=MPITagAllocator.get_tag, init=False)
    __nx: int = field(init=False)
    __ny: int = field(init=False)
    __nz: int = field(init=False)

    def __post_init__(self):
        self.__rank = self.comm.Get_rank()
        self.__commsize = self.comm.Get_size()
        self.__nx, self.__ny, self.__nz = mpi_subdomains_decomposition(self.__commsize)
        ix = self.__rank % self.__nx
        iy = (self.__rank // self.__nx) % self.__ny
        iz = self.__rank // (self.__nx * self.__ny)
        self._domain_index = (ix, iy, iz)
        self.file = None
        if self.__rank == 0:
            try:
                self.file = open(self.file_path, "r", encoding=self.encoding)
            except Exception:
                self._handle_exception()

    def __enter__(self) -> "LAMMPSReaderMPI":
        return self

    def __del__(self) -> None:
        if self.__rank == 0 and self.file is not None:
            self.file.close()

    def __exit__(
        self,
        exc_type: None | type[BaseException] = None,
        exc_value: None | BaseException = None,
        exc_traceback: None | TracebackType = None,
    ) -> bool:
        """Exits the context manager."""
        if self.__rank == 0 and self.file is not None:
            self.file.close()
        return False

    def __get_dtype(
        self, file: TextIO
    ) -> Tuple[list[str], list[Type[int | float]], np.dtype]:
        items = file.readline().split()[2:]
        types = [
            np.int64 if it in ("id", "type", "element", "size") else np.float64
            for it in items
        ]
        if all(c in items for c in ("x", "y", "z")):
            items += ["xs", "ys", "zs"]
            types += [np.float64] * 3
        return items, types, np.dtype(list(zip(items, types)))

    def __process_header(self, file: TextIO) -> Dict[str, Any]:
        data = defaultdict(None)
        line = file.readline()
        if not line:
            return {}
        if line.strip() == "ITEM: TIME":
            data["time"] = float(file.readline())
            file.readline()
        else:
            # rewind if no time item
            file.seek(file.tell() - len(line))
        data["timestep"] = int(file.readline())
        file.readline()
        data["natoms"] = int(file.readline())
        data["boundary"] = file.readline().split()[3:]
        bounds = []
        for _ in range(3):
            b = file.readline().split()
            bounds.append(b)
        data["xlo"], data["xhi"] = map(float, bounds[0][:2])
        data["ylo"], data["yhi"] = map(float, bounds[1][:2])
        data["zlo"], data["zhi"] = map(float, bounds[2][:2])
        return data

    @mpi_safe_method
    def __iter__(self) -> Generator[Tuple[Dict[str, Any], npt.NDArray], None, None]:
        while True:
            # header broadcast
            data = self.comm.bcast(
                self.__process_header(self.file) if self.__rank == 0 else None, root=0
            )
            if data is None or not data:
                break
            # dtype broadcast
            items, types, dtype = self.comm.bcast(
                self.__get_dtype(self.file) if self.__rank == 0 else (None, None, None),
                root=0,
            )
            data.update({"items": items, "types": types, "dtype": dtype})

            # calculate raw line counts
            natoms = data["natoms"]
            counts = [
                (natoms // self.__commsize)
                + (1 if i < (natoms % self.__commsize) else 0)
                for i in range(self.__commsize)
            ]

            # distribute chunks
            if self.__rank == 0:
                chunks = []
                for cnt in counts:
                    chunk = [self.file.readline().split() for _ in range(cnt)]
                    chunks.append(chunk)
                raw = chunks[0]
                for r in range(1, self.__commsize):
                    self.comm.send(chunks[r], dest=r, tag=self.__comm_tag)
            else:
                raw = self.comm.recv(source=0, tag=self.__comm_tag)

            # build structured array
            arr = np.empty(len(raw), dtype=dtype)
            for i, fields in enumerate(raw):
                for j, key in enumerate(items):
                    if key in ("xs", "ys", "zs"):
                        continue
                    arr[key][i] = types[j](fields[j])

            # subdomain info: indices and physical bounds
            xlo, xhi = data["xlo"], data["xhi"]
            ylo, yhi = data["ylo"], data["yhi"]
            zlo, zhi = data["zlo"], data["zhi"]
            dx = (xhi - xlo) / self.__nx
            dy = (yhi - ylo) / self.__ny
            dz = (zhi - zlo) / self.__nz
            ix, iy, iz = self._domain_index
            data["subdomain_index"] = (ix, iy, iz)
            data["subdomain_bounds"] = {
                "xlo": xlo + ix * dx,
                "xhi": xlo + (ix + 1) * dx,
                "ylo": ylo + iy * dy,
                "yhi": ylo + (iy + 1) * dy,
                "zlo": zlo + iz * dz,
                "zhi": zlo + (iz + 1) * dz,
            }
            # normalize scaled coordinates based on true positions
            if all(c in items for c in ("xs", "ys", "zs")):
                arr["xs"] = (arr["x"] - xlo) / (xhi - xlo)
                arr["ys"] = (arr["y"] - ylo) / (yhi - ylo)
                arr["zs"] = (arr["z"] - zlo) / (zhi - zlo)
            # attach atoms
            data["atoms"] = arr
            data["natoms"] = len(arr)
            yield data

        if self.file:
            self.file.close()

    @mpi_safe_method
    def close(self) -> None:
        """Closes the file associated with this writer."""
        if self.__rank == 0 and not self.file.closed:
            self.file.close()



"""dpa analysis module."""

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D
from numpy import typing as npt

from irradiapy import materials, utils
from irradiapy.io import LAMMPSReader
from irradiapy.srim import SRIMDB

# region Histograms


# NOTE: The first argument type will determine the calculation method when other BCA methods are
# implemented.
# NOTE: SRIM support only for monoelemental targets.
def get_dpa_1d(
    srimdb: SRIMDB,
    path_debris: Path,
    path_db: Path,
    fluence: float,
    axis: str = "x",
    nbins: int = 100,
    depth_offset: float = 0.0,
) -> None:
    """Perform a dpa histogram as a function of depth along a specified axis.

    Note
    ----
    dpa is calculated from vacancies in the debris file.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    path_debris : Path
        File containing all collision debris to analyze.
    path_db : Path
        Path to the database file.
    fluence : float
        Fluence, in ions/A2.
    axis : str, optional (default="x")
        Axis along which the histogram was computed. It can be `"x"`, `"y"`, or `"z"`.
    nbins : int, optional (default=100)
        Number of depth bins. Depth range determined from `path_debris` defect positions.
    depth_offset : float, optional (default=0.0)
        Offset to add to the depth values.
    """

    # Get materials
    atomic_number1 = list(
        srimdb.trimdat.read(what="atom_numb", condition="WHERE ion_numb = 1")
    )[0][0]
    mat_pka = materials.MATERIALS_BY_ATOMIC_NUMBER[atomic_number1]
    atomic_number2 = srimdb.target.layers[0].elements[0].atomic_number
    mat_target = materials.MATERIALS_BY_ATOMIC_NUMBER[atomic_number2]
    nions = srimdb.nions

    # SRIM COLLISON.txt dpa
    depths, defects_nrt, defects_fer_arc = [], [], []
    what = "depth, recoil_energy" if axis == "x" else f"{axis}, recoil_energy"
    for depth, pka_e in srimdb.collision.read(what=what):
        depths.append(depth)
        tdam = mat_target.epka_to_tdam(mat_pka, pka_e)
        defects_nrt.append(mat_target.calc_nrt_dpa(tdam))
        defects_fer_arc.append(mat_target.calc_fer_arc_dpa(tdam))
    depths = np.array(depths)
    defects_nrt = np.array(defects_nrt)
    defects_fer_arc = np.array(defects_fer_arc)

    # Debris dpa
    depth_debris = np.array([], dtype=np.float64)
    for data_defects in LAMMPSReader(path_debris):
        vacs = data_defects["atoms"][data_defects["atoms"]["type"] == 0]
        depth_debris = np.concatenate((depth_debris, vacs[axis]))

    # Depth binning (use debris to set bin edges)
    depth_edges = np.histogram_bin_edges(depth_debris, bins=nbins)
    depth_centers = (depth_edges[1:] - depth_edges[:-1]) / 2.0 + depth_edges[:-1]
    width = depth_edges[1] - depth_edges[0]
    srim_digitize = np.digitize(depths, depth_edges)

    # NRT-dpa
    hist_nrt = np.array(
        [np.sum(defects_nrt[srim_digitize == i]) for i in range(1, nbins + 1)]
    )
    dpa_nrt = hist_nrt / nions * fluence / mat_target.density / width
    # fer-arc-dpa
    hist_fer_arc = np.array(
        [np.sum(defects_fer_arc[srim_digitize == i]) for i in range(1, nbins + 1)]
    )
    dpa_fer_arc = hist_fer_arc / nions * fluence / mat_target.density / width
    # debris-dpa
    hist_debris, _ = np.histogram(depth_debris, bins=depth_edges)
    dpa_debris = hist_debris / nions * fluence / mat_target.density / width

    # Save to database
    utils.sqlite.insert_array(
        path_db,
        f"dpa_1D_{axis}",
        depth_centers=depth_centers + depth_offset,
        dpa_nrt=dpa_nrt,
        dpa_fer_arc=dpa_fer_arc,
        dpa_debris=dpa_debris,
    )


def read_dpa_1d(path_db: Path, axis: str = "x") -> dict[str, npt.NDArray[np.float64]]:
    """Read the 1D dpa histogram from the database.

    Parameters
    ----------
    path_db : Path
        Path to the SQLite database file.
    axis : str, optional (default="x")
        Axis along which the histogram was computed. It can be `"x"`, `"y"`, or `"z"`.

    Returns
    -------
    dict[str, npt.NDArray[np.float64]]
        A dictionary containing the 1D dpa histogram data.
    """
    data = utils.sqlite.read_array(path_db, f"dpa_1D_{axis}")
    return data


# region Plots


def plot_dpa_1d(
    path_db: Path,
    path_plot: None | Path = None,
    path_fit: None | Path = None,
    axis: str = "x",
    depth_offset: float = 0.0,
    p0: None | float = None,
    asymmetry: float = 1.0,
    dpi: int = 300,
) -> None:
    """Plot the 1D dpa analysis results from the database.

    Parameters
    ----------
    path_db : Path
        Path to the database file.
    path_plot : Path, optional (default=None)
        Output path for the plot. If None, the plot is shown.
    path_fit : Path, optional (default=None)
        Output path for the fit parameters.
    axis : str, optional (default="x")
        Axis along which the histogram was computed. It can be `"x"`, `"y"`, or `"z"`.
    depth_offset : float, optional (default=0.0)
        Offset to add to the depth values.
    p0 : float, optional (default=None)
        Initial guess of fit parameters.
    asymmetry : float, optional (default=1.0)
        Asymmetry fit parameter bound.
    dpi : int, optional (default=300)
        Dots per inch.
    """

    data = read_dpa_1d(path_db, axis=axis)
    depth_centers = data["depth_centers"] + depth_offset
    dpa_nrt = data["dpa_nrt"]
    dpa_fer_arc = data["dpa_fer_arc"]
    dpa_debris = data["dpa_debris"]

    total_nrt = dpa_nrt.sum()
    total_debris = dpa_debris.sum()

    # Fit dpa_debris
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec()
    ax = fig.add_subplot(gs[0, 0])
    # Scatter
    ax.set_xlabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_ylabel("dpa")
    ax.scatter(depth_centers, dpa_nrt, label="NRT-dpa")
    ax.scatter(depth_centers, dpa_fer_arc, label="fer-arc-dpa")
    ax.scatter(depth_centers, dpa_debris, label="debris-dpa")
    efficiency = [
        Line2D(
            [0],
            [0],
            color="none",
            label=r"$\overline{\xi}$ = "
            + rf"{round(total_debris / total_nrt * 100)} %",
        )
    ]

    if path_fit:
        try:
            popt, _, dpa_fit = utils.math.fit_lorentzian(
                depth_centers, dpa_debris, p0, asymmetry
            )
            if path_fit:
                with open(path_fit, "w", encoding="utf-8") as file:
                    file.write("Debris dpa lorentzian fit: z0, sigma, A, a\n")
                    file.write(
                        (
                            "See Eq. (1) of Nuclear Instruments and Methods in Physics "
                            "Research B 500-501 (2021) 52-56\n"
                        )
                    )
                    file.write(", ".join(map(str, popt)) + "\n")
            ax.plot(
                depth_centers,
                dpa_fit(depth_centers),
                label="debris-dpa fit",
            )
        except Exception as exc:  # pylint: disable=broad-except
            print(f"Fit failed: {exc}")

    ax.legend(handles=ax.get_legend_handles_labels()[0] + efficiency)
    fig.tight_layout()
    if path_plot:
        plt.savefig(path_plot, dpi=dpi)
    else:
        plt.show()
    plt.close()



"""Utilities for MPI-related functionalities."""

# pylint: disable=no-name-in-module, broad-except

import sys
import traceback
from functools import wraps
from pathlib import Path

from mpi4py import MPI

from irradiapy.utils.math import repeated_prime_factors


def broadcast_variables(root: int, comm: MPI.Comm, *variables) -> list:
    """Broadcasts variables.

    Parameters
    ----------
    root : int
        The rank of the process that will broadcast the variables.
    comm : MPI.Comm
        The MPI communicator.
    variables : tuple
        The variables to broadcast.

    Returns
    -------
    list
        The broadcasted variables.
    """
    return [comm.bcast(var, root=root) for var in variables]


def rm_file(path_file: Path, comm: MPI.Comm) -> None:
    """Remove a file."""
    rank = comm.Get_rank()
    if rank == 0:
        path_file.unlink(missing_ok=True)
    comm.Barrier()


def cp_file(original: Path, target: Path, comm: MPI.Comm) -> None:
    """Copy a file from `original` to `target`, overwriting `target` if it exists."""
    rank = comm.Get_rank()
    if rank == 0:
        print(f"\n\n\nCopying {original} to {target}")
        target.write_bytes(original.read_bytes())
    comm.Barrier()


def mv_file(original: Path, target: Path, comm: MPI.Comm) -> None:
    """Move a file from `original` to `target`, overwriting `target` if it exists."""
    rank = comm.Get_rank()
    if rank == 0:
        target.write_bytes(original.read_bytes())
        original.unlink()
    comm.Barrier()


def ap_rm_file(original: Path, target: Path, comm: MPI.Comm) -> None:
    """Append content from `original` to `target` and delete `original`."""
    rank = comm.Get_rank()
    if rank == 0:
        original_content = original.read_bytes()
        with target.open("ab") as f:
            f.write(original_content)
        original.unlink()
    comm.Barrier()


def mpi_safe_method(method):
    """Decorator that wraps an MPI-using method so any exception prints a
    traceback with the current rank and then calls MPI.Abort.

    The method should be a member of a class that has `comm` and `rank` attributes.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except Exception:
            tb = traceback.format_exc()
            sys.stderr.write(f"Rank {self.rank} raised an exception:\n{tb}\n")
            sys.stderr.flush()
            self.comm.Abort(1)

    return wrapper


def mpi_subdomains_decomposition(n: int) -> tuple[int, int, int]:
    """Factor `n` into three integers `nx`, `ny`, `nz` for MPI decomposition into subdomains.

    `nx`, `ny` and `nz` are such that:
    - `nx * ny * nz == n`
    - the maximum of (`nx`, `ny`, `nz`) divided by the minimum is as small as possible.

    Parameters
    ----------
    n : int
        The number of processes to decompose into subdomains.

    Returns
    -------
    tuple[int, int, int]
        A tuple containing the number of subdomains in the x, y, and z directions.
    """
    # Start with 1,1,n
    dims = [1, 1, 1]
    facs = repeated_prime_factors(n)
    # For each prime, multiply into the currently smallest dimension
    for p in sorted(facs, reverse=True):
        # Sort so that dims[0] ≤ dims[1] ≤ dims[2]
        dims.sort()
        dims[0] *= p
    return tuple(sorted(dims))


class MPIExceptionHandlerMixin:
    """Provides a common MPI exception handler method when `mpi_safe_method` cannot be used.

    This is useful for `__init__` and `__post_init__` methods where decorators cannot be applied
    because `self.comm` and `self.rank` are not yet defined before the function is called. You can
    use this method after the `comm` and `rank` attributes are initialized in a try/except block.
    """

    def _handle_exception(self) -> None:
        tb = traceback.format_exc()
        sys.stderr.write(f"[Rank {self.rank}] Exception:\n{tb}\n")
        sys.stderr.flush()
        self.comm.Abort(1)


class MPITagAllocator:
    """A class to allocate unique tags for processes."""

    _next_tag = 0

    @classmethod
    def get_tag(cls):
        """Get a unique tag for the current process.

        Warning
        -------
        This method should not be called in a statement that is only executed by one rank,
        such as in `if rank == 0:`. It is designed to be called by all ranks
        to ensure that all ranks receive the same tag.
        """
        tag = cls._next_tag
        cls._next_tag += 1
        return tag



"""Fe material."""

from math import sqrt

from irradiapy.materials.material import Material
from irradiapy.srim.target import element as srim_element

Fe = Material(
    atomic_number=26,
    mass_number=55.845,
    a0=2.87,
    cutoff_sia=2.87 * sqrt(2.0),
    cutoff_vac=2.87,
    dist_fp=4.0 * 2.87,
    density=8.5e-2,
    ed_min=20,
    ed_avr=40,
    b_arc=-0.568,
    c_arc=0.286,
    srim_element=srim_element.Element(
        symbol="Fe",
        atomic_number=26,
        atomic_mass=55.85,
        e_d=40.0,
        e_l=5.8,
        e_s=4.34,
        density=7.8658,
    ),
)



"""This module contains the `LAMMPSReader` class."""

from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Generator, TextIO, Type

import numpy as np


@dataclass
class LAMMPSReader:
    """A class to read data from a LAMMPS dump file.

    Note
    ----
    Assumed orthogonal simulation box.

    Attributes
    ----------
    file_path : Path
        The path to the LAMMPS dump file.
    encoding : str, optional (default="utf-8")
        The file encoding.
    """

    file_path: Path
    encoding: str = "utf-8"
    file: TextIO = field(default=None, init=False)

    def __post_init__(self) -> None:
        self.file = open(self.file_path, encoding=self.encoding)

    def __del__(self) -> None:
        if self.file is not None:
            self.file.close()

    def __iter__(
        self,
    ) -> Generator[
        dict,  # Changed from tuple[...] to dict
        None,
        None,
    ]:
        """Read the file as an iterator, timestep by timestep.

        Yields
        ------
        dict
            A dictionary containing the timestep data with keys:
            'time' (optional), 'timestep', 'natoms', 'boundary', 'xlo', 'xhi',
            'ylo', 'yhi', 'zlo', 'zhi', and 'atoms' (as a numpy structured array).
        """
        while True:
            data = defaultdict(None)
            line = self.file.readline()
            if not line:
                break
            if line == "ITEM: TIME\n":
                data["time"] = float(self.file.readline())
                self.file.readline()
            data["timestep"] = int(self.file.readline())
            self.file.readline()
            data["natoms"] = int(self.file.readline())
            data["boundary"] = self.file.readline().split()[-3:]
            data["xlo"], data["xhi"] = map(float, self.file.readline().split())
            data["ylo"], data["yhi"] = map(float, self.file.readline().split())
            data["zlo"], data["zhi"] = map(float, self.file.readline().split())

            line = self.file.readline()
            items, types, dtype = self.__get_dtype(line)

            data["atoms"] = np.empty(data["natoms"], dtype=dtype)
            for i in range(data["natoms"]):
                line = self.file.readline().split()
                for j, item in enumerate(items):
                    data["atoms"][i][item] = types[j](line[j])
            yield data

        self.file.close()

    def __get_dtype(
        self, line: str
    ) -> tuple[list[str], list[Type[int | float]], np.dtype]:
        """Get the data type of the simulation data.

        Parameters
        ----------
        line : str
            The line containing the data type.

        Returns
        -------
        tuple[list[str], list[Type[int | float]], np.dtype]
            The names of the data items, the types of the data items,
            and the data type.
        """
        items = line.split()[2:]
        types = [
            int if item in ("id", "type", "element", "size") else float
            for item in items
        ]
        dtype = np.dtype([(item, type) for item, type in zip(items, types)])
        return items, types, dtype

    def close(self) -> None:
        """Close the file associated with this reader."""
        self.file.close()



"""This module contains the `LAMMPSWriter` class."""

from dataclasses import dataclass, field
from pathlib import Path
from types import TracebackType
from typing import TextIO

from irradiapy import config


@dataclass
class LAMMPSWriter:
    """A class to write data like a LAMMPS dump file.

    Attributes
    ----------
    file_path : Path
        The path to the LAMMPS dump file.
    mode : str, optional (default="w")
        The file open mode.
    excluded_items : list[str], optional (default=irradiapy.config.EXCLUDED_ITEMS)
        Atom fields to exclude from output.
    encoding : str, optional (default=irradiapy.config.ENCODING)
        The file encoding.
    int_format : str, optional (default=irradiapy.config.INT_FORMAT)
        The format for integers.
    float_format : str, optional (default=irradiapy.config.FLOAT_FORMAT)
        The format for floats.
    """

    file_path: Path
    mode: str = "w"
    excluded_items: list[str] = field(default_factory=lambda: config.EXCLUDED_ITEMS)
    encoding: str = field(default_factory=lambda: config.ENCODING)
    int_format: str = field(default_factory=lambda: config.INT_FORMAT)
    float_format: str = field(default_factory=lambda: config.FLOAT_FORMAT)
    file: TextIO = field(default=None, init=False)

    def __post_init__(self) -> None:
        self.file = open(self.file_path, self.mode, encoding=self.encoding)

    def __enter__(self) -> "LAMMPSWriter":
        return self

    def __del__(self) -> None:
        if self.file is not None:
            self.file.close()

    def __exit__(
        self,
        exc_type: None | type[BaseException] = None,
        exc_value: None | BaseException = None,
        exc_traceback: None | TracebackType = None,
    ) -> bool:
        if self.file is not None:
            self.file.close()
        return False

    def close(self) -> None:
        """Closes the file associated with this writer."""
        if self.file is not None:
            self.file.close()

    def write(self, data: dict) -> None:
        """Writes the data (from LAMMPSReader/BZIP2LAMMPSReader) to the file.

        Parameters
        ----------
        data : dict
            The dictionary containing the timestep data.
        """
        if data.get("time") is not None:
            self.file.write(f"ITEM: TIME\n{data['time']}\n")
        self.file.write(f"ITEM: TIMESTEP\n{data['timestep']}\n")
        self.file.write(f"ITEM: NUMBER OF ATOMS\n{data['natoms']}\n")
        self.file.write(f"ITEM: BOX BOUNDS {' '.join(data['boundary'])}\n")
        self.file.write(
            f"{self.float_format % data['xlo']} {self.float_format % data['xhi']}\n"
        )
        self.file.write(
            f"{self.float_format % data['ylo']} {self.float_format % data['yhi']}\n"
        )
        self.file.write(
            f"{self.float_format % data['zlo']} {self.float_format % data['zhi']}\n"
        )

        atoms = data["atoms"]
        field_names = [f for f in atoms.dtype.names if f not in self.excluded_items]
        self.file.write(f"ITEM: ATOMS {' '.join(field_names)}\n")

        formatters = []
        for field_name in field_names:
            dtype = atoms.dtype[field_name]
            if dtype.kind == "i":
                formatters.append(self.int_format)
            elif dtype.kind == "f":
                formatters.append(self.float_format)
            else:
                formatters.append("%s")

        for row in atoms:
            self.file.write(
                " ".join(
                    fmt % row[field_name]
                    for fmt, field_name in zip(formatters, field_names)
                )
                + "\n"
            )



"""Cu material."""

from math import sqrt

from irradiapy.materials.material import Material
from irradiapy.srim.target import element as srim_element

Cu = Material(
    atomic_number=29,
    mass_number=63.546,
    a0=3.61491,
    cutoff_sia=3.61491 * sqrt(2.0),
    cutoff_vac=3.61491,
    dist_fp=4.0 * 3.61491,
    density=8.46e-2,
    ed_min=25,
    ed_avr=33,
    b_arc=-0.68,
    c_arc=0.16,
    srim_element=srim_element.Element(
        symbol="Cu",
        atomic_number=29,
        atomic_mass=63.546,
        e_d=33.0,
        e_l=4.4,
        e_s=3.52,
        density=8.92,
    ),
)



"""Utility functions related to SRIM data analysis and debris production."""

from collections import defaultdict
from itertools import combinations
from pathlib import Path
from typing import Callable

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.image import NonUniformImage

from irradiapy import dtypes, materials
from irradiapy.damagedb import DamageDB
from irradiapy.io.lammpswriter import LAMMPSWriter
from irradiapy.srim.srimdb import SRIMDB
from irradiapy.utils.math import fit_gaussian, fit_scaling_law

# region range3d


def plot_injected(
    srimdb: SRIMDB,
    bins: int = 100,
    plot_path: None | Path = None,
    path_fit: None | Path = None,
    p0: None | float = None,
    asymmetry: float = 1.0,
    dpi: int = 300,
) -> None:
    """Plot injected ions final depth distribution.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    bins : int, optional (default=100)
        Depth bins.
    plot_path : Path, optional (default=None)
        Output path for the plot. If `None`, the plot is shown.
    path_fit : Path, optional (default=None)
        Output path for the fit parameters.
    p0 : float, optional (default=None)
        Initial guess of fit parameters.
    asymmetry : float, optional (default=1.0)
        Asymmetry fit parameter bound.
    dpi : int, optional (default=300)
        Dots per inch.
    """
    # Read
    depths = np.array([ion[0] for ion in srimdb.range3d.read(what="depth")])
    # Histogram
    histogram, depth_edges = np.histogram(depths, bins=bins)
    depth_centers = (depth_edges[:-1] + depth_edges[1:]) / 2.0
    # Fit
    fit, injected_fit = False, None
    if path_fit:
        try:
            popt, _, injected_fit = fit_gaussian(
                depth_centers, histogram, p0, asymmetry
            )
            if path_fit:
                with open(path_fit, "w", encoding="utf-8") as file:
                    file.write("Injected atoms gaussian fit: z0, sigma, A, a\n")
                    file.write(
                        (
                            "See Eq. (1) of Nuclear Instruments and Methods in Physics "
                            "Research B 500-501 (2021) 52-56\n"
                        )
                    )
                    file.write(", ".join(map(str, popt)) + "\n")
            fit = True
        except Exception as exc:  # pylint: disable=broad-except
            print(f"Fit failed: {exc}")
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec()
    # Scatter
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_ylabel("Counts per ion")
    ax.scatter(depth_centers, histogram)
    if fit:
        ax.plot(
            depth_centers,
            injected_fit(depth_centers),
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0],
        )
    # Finish
    fig.tight_layout()
    if plot_path:
        plt.savefig(plot_path, dpi=dpi)
    else:
        plt.show()
    plt.close()


# region collision


def plot_pka_distribution(
    srimdb: SRIMDB,
    bins: int = 100,
    plot_path: None | Path = None,
    fit_path: None | Path = None,
    dpi: int = 300,
) -> Callable:
    """Plot the PKA energy distribution and tries to fit it.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    bins : int, optional (default=100)
        Energy bins. The fit will be done over non-empty bins.
    plot_path : Path, optional (default=None)
        Output path for the plot. If `None`, the plot is shown.
    fit_path : Path, optional (default=None)
        Output path for the fit parameters.
    dpi : int, optional (default=300)
        Dots per inch.

    Returns
    -------
    Callable
        Scaling law function.
    """
    # Read
    nions = srimdb.nions
    pka_es = np.array(list(srimdb.collision.read(what="recoil_energy")))
    # Histogram
    pka_hist, pka_edges = np.histogram(pka_es, bins=bins)
    pka_hist = pka_hist / nions
    pka_centers = pka_edges[:-1] + (pka_edges[1:] - pka_edges[:-1]) / 2.0
    # Fit using fit_scaling_law
    fit = False
    try:
        mask = pka_hist > 0
        a, s, curve = fit_scaling_law(pka_centers[mask] / 1e3, pka_hist[mask])
        fit = True
    except Exception as exc:  # pylint: disable=broad-except
        print(f"Fit failed: {exc}")
    if fit and fit_path:
        with open(fit_path, "w", encoding="utf-8") as file:
            file.write(f"PKA energy scaling law (x in eV)\nA/x**S\nA S\n{a}, {s}\n")
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec()
    # Scatter
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(r"$E_{PKA}$ (keV)")
    ax.set_ylabel("Counts per ion")
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.scatter(pka_centers / 1e3, pka_hist)
    if fit:
        ax.plot(
            pka_centers / 1e3,
            curve(pka_centers / 1e3),
            color=plt.rcParams["axes.prop_cycle"].by_key()["color"][0],
        )
    # Finish
    fig.tight_layout()
    if plot_path:
        plt.savefig(plot_path, dpi=dpi)
    else:
        plt.show()
    plt.close()
    return curve


def plot_energy_depth(
    srimdb: SRIMDB,
    depth_bins: int = 100,
    pka_ebins: int = 100,
    pka_e_max: float = 200,
    plot_high_path: None | Path = None,
    plot_low_path: None | Path = None,
    dpi: int = 300,
) -> None:
    """Plots the depth-energy distribution of PKAs.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    depth_bins : int, optional (default=100)
        Number of bins for depth histogram.
    pka_ebins : int, optional (default=100)
        Number of bins for PKA energy histogram.
    pka_e_max : float, optional (default=200)
        Maximum PKA energy.
    plot_high_path : Path, optional (default=None)
        Output path for the high energy plot. If `None`, the plot is shown.
    plot_low_path : Path, optional (default=None)
        Output path for the low energy plot. If `None`, the plot is shown.
    dpi : int, optional (default=300)
        Dots per inch for the plot.
    """
    # Read
    nions = srimdb.nions
    data = np.array(list(srimdb.collision.read(what="depth, recoil_energy")))
    depths, pka_es = data[:, 0], data[:, 1]
    # Low energy, linear
    depth_edges = np.histogram_bin_edges(depths, bins=depth_bins)
    pka_e_edges = np.histogram_bin_edges(
        pka_es, bins=pka_ebins, range=(pka_es.min(), pka_e_max)
    )
    hist, _, _ = np.histogram2d(depths, pka_es, bins=[depth_edges, pka_e_edges])
    hist /= nions
    depth_centers = depth_edges[:-1] + (depth_edges[1:] - depth_edges[:-1]) / 2.0
    pka_e_centers = pka_e_edges[:-1] + (pka_e_edges[1:] - pka_e_edges[:-1]) / 2.0
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec(nrows=1, ncols=2, width_ratios=[0.95, 0.05])
    cmap = plt.cm.get_cmap("viridis")
    cmap.set_under(plt.rcParams["axes.facecolor"])
    # Color map
    ax = fig.add_subplot(gs[0, 0])
    ax.set_ylabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_xlabel(r"$E_{PKA}$ (eV) (low energies)")
    ax.set_xlim(pka_e_edges[[0, -1]])
    ax.set_ylim(depth_edges[[0, -1]])
    im = NonUniformImage(
        ax, cmap=cmap, extent=(*pka_e_edges[[0, -1]], *depth_edges[[0, -1]])
    )
    im.set_clim(vmin=1 / nions)
    im.set_data(pka_e_centers, depth_centers, hist)
    ax.add_image(im)
    # Color bar
    cax = fig.add_subplot(gs[0, 1])
    fig.colorbar(im, cax, label="Counts per ion")
    plt.tight_layout()
    if plot_low_path:
        plt.savefig(plot_low_path, dpi=dpi)
    else:
        plt.show()
    plt.close()
    # All energies, log
    depth_edges = np.histogram_bin_edges(depths, bins=depth_bins)
    pka_e_edges = np.histogram_bin_edges(pka_es, bins=pka_ebins)
    hist, _, _ = np.histogram2d(depths, pka_es, bins=[depth_edges, pka_e_edges])
    hist /= nions
    depth_centers = depth_edges[:-1] + (depth_edges[1:] - depth_edges[:-1]) / 2.0
    pka_e_centers = pka_e_edges[:-1] + (pka_e_edges[1:] - pka_e_edges[:-1]) / 2.0
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec(nrows=1, ncols=2, width_ratios=[0.95, 0.05])
    # Color map
    ax = fig.add_subplot(gs[0, 0])
    ax.set_ylabel(r"Depth ($\mathrm{\AA}$)")
    ax.set_xlabel(r"$E_{PKA}$ (keV)")
    ax.set_xlim(pka_e_edges[[0, -1]] / 1e3)
    ax.set_ylim(depth_edges[[0, -1]])
    im = NonUniformImage(
        ax,
        cmap=cmap,
        norm="log",
        extent=(*pka_e_edges[[0, -1]] / 1e3, *depth_edges[[0, -1]]),
    )
    im.set_clim(vmin=1 / nions)
    im.set_data(pka_e_centers / 1e3, depth_centers, hist)
    ax.add_image(im)
    # Color bar
    cax = fig.add_subplot(gs[0, 1])
    fig.colorbar(im, cax, label="Counts per ion")
    # Finish
    plt.tight_layout()
    if plot_high_path:
        plt.savefig(plot_high_path, dpi=dpi)
    else:
        plt.show()
    plt.close()


def plot_distances(
    srimdb: SRIMDB,
    pka_e_lim: float = 5e3,
    dist_bins: int = 100,
    energy_bins: int = 100,
    plot_path: None | Path = None,
    dpi: int = 300,
) -> None:
    """Plots a 2D histogram of pairwise distances and sum of PKA energies for each ion.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    pka_elim : float, optional (default=5e3)
        Minimum recoil energy to consider.
    dist_bins : int, optional (default=100)
        Number of bins for the distance histogram.
    energy_bins : int, optional (default=100)
        Number of bins for the energy histogram.
    plot_path : Path, optional (default=None)
        Output path for the plot. If `None`, the plot is shown.
    dpi : int, optional (default=300)
        Dots per inch for the plot.
    """
    nions = srimdb.nions
    distances = []
    energies = []
    # Get pairwise distances and energies for each ion
    for nion in range(1, nions + 1):
        data = np.array(
            list(
                srimdb.collision.read(
                    what="depth, y, z, recoil_energy",
                    condition=f"WHERE ion_numb = {nion} AND recoil_energy >= {pka_e_lim}",
                )
            )
        )
        if len(data):
            pos = data[:, :3]
            pka_e = data[:, 3]
            for i, j in combinations(range(len(pos)), 2):
                distance = np.linalg.norm(pos[i] - pos[j])
                energy = pka_e[i] + pka_e[j]
                distances.append(distance)
                energies.append(energy)
    distances = np.array(distances)
    energies = np.array(energies) / 1e3
    # Histogram
    dist_edges = np.histogram_bin_edges(distances, bins=dist_bins)
    energies_edges = np.histogram_bin_edges(energies, bins=energy_bins)
    hist, _, _ = np.histogram2d(distances, energies, bins=[dist_edges, energies_edges])
    hist /= nions
    dist_centers = dist_edges[:-1] + (dist_edges[1:] - dist_edges[:-1]) / 2.0
    energies_centers = (
        energies_edges[:-1] + (energies_edges[1:] - energies_edges[:-1]) / 2.0
    )
    # Plot
    fig = plt.figure()
    gs = fig.add_gridspec(nrows=1, ncols=2, width_ratios=[0.95, 0.05])
    cmap = plt.cm.get_cmap("viridis")
    cmap.set_under(plt.rcParams["axes.facecolor"])
    # Color map
    ax = fig.add_subplot(gs[0, 0])
    ax.set_xlabel(r"Pairwise distance ($\mathrm{\AA}$)")
    ax.set_ylabel(r"Sum of $E_{PKA}$ (keV)")
    ax.set_ylim(energies_edges[[0, -1]])
    ax.set_xlim(dist_edges[[0, -1]])
    im = NonUniformImage(
        ax,
        cmap=cmap,
        extent=(*dist_edges[[0, -1]], *energies_edges[[0, -1]]),
    )
    im.set_clim(vmin=1 / nions)
    im.set_data(dist_centers, energies_centers, hist)
    ax.add_image(im)
    # Color bar
    cax = fig.add_subplot(gs[0, 1])
    fig.colorbar(im, cax, label="Counts per ion")
    # Finish
    fig.tight_layout()
    if plot_path:
        plt.savefig(plot_path, dpi=dpi)
    else:
        plt.show()
    plt.close()


# region Debris


def generate_debris(
    srimdb: SRIMDB,
    dir_mddb: Path,
    compute_tdam: bool,
    path_debris: Path,
    tdam_mode: "materials.Material.TdamMode",
    dpa_mode: "materials.Material.DpaMode",
    add_injected: bool,
    outsiders: bool,
    seed: None | int = None,
    depth_offset: None | float = 0.0,
    ylo: None | float = None,
    yhi: None | float = None,
    zlo: None | float = None,
    zhi: None | float = None,
) -> None:
    """Turns SRIM's collisions into `.xyz` files for the given database of cascades' debris.

    Warning
    -------
        Assumes a monolayer monoatomic target and same element for all ions.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    dir_mddb : Path
        Directory where the database of cascades' debris is stored.
    compute_tdam : bool
        Whether to transform the PKA energies into damage energies. It should be `True` for
        MD simulations without electronic stopping.
    path_debris : Path
        Directory where the ions debris will be stored as `.xyz` files.
    tdam_mode : materials.Material.TdamMode
        Mode to convert the PKA energy into damage energy.
    dpa_mode : materials.Material.DpaMode
        Formula to convert the residual energy into Frenkel pairs.
    add_injected : bool
        Whether to add the injected interstitial.
    outsiders : bool
        Whether to include defects generated outside the material (PKAs close to surfaces).
        This can cause an inmbalance between the number of vacancies and interstitials.
    seed : int, optional
        Random seed for reproducibility.
    depth_offset : float, optional (default=0.0)
        Offset to add to the depth of the defects.
    ylo : float, optional (default=None)
        Minimum y-coordinate for the defects. If `None`, will be set to minus the target width.
    yhi : float, optional (default=None)
        Maximum y-coordinate for the defects. If `None`, will be set to the target width.
    zlo : float, optional (default=None)
        Minimum z-coordinate for the defects. If `None`, will be set to minus the target width.
    zhi : float, optional (default=None)
        Maximum z-coordinate for the defects. If `None`, will be set to the target width.
    """
    xlo, xhi = depth_offset, srimdb.target.layers[0].width + depth_offset
    if ylo is None:
        ylo = -xhi
    if yhi is None:
        yhi = xhi
    if zlo is None:
        zlo = -xhi
    if zhi is None:
        zhi = xhi
    pka_atomic_number = next(
        iter(srimdb.trimdat.read(what="atom_numb", condition="WHERE ion_numb = 1"))
    )[0]
    mat_pka = materials.MATERIALS_BY_ATOMIC_NUMBER[pka_atomic_number]
    mat_atomic_number = srimdb.target.layers[0].elements[0].atomic_number
    mat_target = materials.MATERIALS_BY_ATOMIC_NUMBER[mat_atomic_number]
    nions = srimdb.nions
    damagedb = DamageDB(
        dir_mddb=dir_mddb,
        compute_tdam=compute_tdam,
        mat_pka=mat_pka,
        mat_target=mat_target,
        tdam_mode=tdam_mode,
        dpa_mode=dpa_mode,
        seed=seed,
    )
    with LAMMPSWriter(path_debris) as writer:
        for nion in range(1, nions + 1):
            defects = __generate_ion_defects(srimdb, nion, damagedb, add_injected)

            # Apply offsets and cuts
            defects["x"] += depth_offset
            if not outsiders:
                defects = defects[
                    (defects["x"] >= xlo)
                    & (defects["x"] <= xhi)
                    & (defects["y"] >= ylo)
                    & (defects["y"] <= yhi)
                    & (defects["z"] >= zlo)
                    & (defects["z"] <= zhi)
                ]

            data_defects = defaultdict(None)
            data_defects["time"] = 0.0
            data_defects["timestep"] = 0
            data_defects["natoms"] = defects.size
            data_defects["boundary"] = ["ff", "ff", "ff"]
            data_defects["xlo"] = xlo
            data_defects["xhi"] = xhi
            data_defects["ylo"] = ylo
            data_defects["yhi"] = yhi
            data_defects["zlo"] = zlo
            data_defects["zhi"] = zhi
            data_defects["atoms"] = defects
            writer.write(data_defects)


def __generate_ion_defects(
    srimdb: SRIMDB,
    nion: int,
    damagedb: DamageDB,
    add_injected: bool,
) -> dtypes.Defect:
    """Generates the defects for a specific ion in the SRIM simulation.

    Parameters
    ----------
    srimdb : SRIMDB
        SRIM database class.
    nion : int
        Ion number.
    damagedb : DamageDB
        DamageDB class that will choose MD debris.
    add_injected : bool
        Whether to add the injected interstitial.

    Returns
    -------
    dtypes.Defect
        An array containing the defects generated by a single ion.
    """
    defects = np.empty(0, dtype=dtypes.defect)
    for depth, y, z, cosx, cosy, cosz, pka_e in srimdb.collision.read(
        what="depth, y, z, cosx, cosy, cosz, recoil_energy",
        condition=f"WHERE ion_numb = {nion}",
    ):
        pka_pos = np.array([depth, y, z])
        pka_dir = np.array([cosx, cosy, cosz])
        defects_ = damagedb.get_pka_debris(
            pka_e=pka_e, pka_pos=pka_pos, pka_dir=pka_dir
        )
        defects = np.concatenate((defects, defects_))

    if add_injected:
        injected = list(
            srimdb.range3d.read(
                what="depth, y, z", condition=f"WHERE ion_numb = {nion} LIMIT 1"
            )
        )
        # No backscattered or transmitted ion
        if injected:
            atom_number = list(
                srimdb.trimdat.read(
                    what="atom_numb", condition=f"WHERE ion_numb = {nion}"
                )
            )[0][0]
            injected = np.array(
                [(atom_number, injected[0][0], injected[0][1], injected[0][2])],
                dtype=dtypes.defect,
            )
            defects = np.concatenate((defects, injected))
    return defects



"""W material."""

from math import sqrt

from irradiapy.materials.material import Material
from irradiapy.srim.target import element as srim_element

W = Material(
    atomic_number=74,
    mass_number=183.84,
    a0=3.1652,
    cutoff_sia=3.1652 * sqrt(2.0),
    cutoff_vac=3.1652,
    dist_fp=4.0 * 3.1652,
    density=6.3e-2,
    ed_min=42,
    ed_avr=70,
    b_arc=-0.56,
    c_arc=0.12,
    srim_element=srim_element.Element(
        symbol="W",
        atomic_number=74,
        atomic_mass=183.84,
        e_d=70.0,
        e_l=13.2,
        e_s=8.68,
        density=19.3,
    ),
)



"""Numpy structured array dtypes."""

from typing import Annotated, Any

import numpy as np
from numpy import typing as npt

# region General
atom = np.dtype(
    [
        ("type", np.int64),
        ("x", np.float64),
        ("y", np.float64),
        ("z", np.float64),
    ]
)
Atom = Annotated[npt.NDArray[Any], atom]
defect = np.dtype(
    [
        ("type", np.int64),
        ("x", np.float64),
        ("y", np.float64),
        ("z", np.float64),
    ]
)
Defect = Annotated[npt.NDArray[Any], defect]
acluster = np.dtype(
    [
        ("type", np.int64),
        ("x", np.float64),
        ("y", np.float64),
        ("z", np.float64),
        ("cluster", np.int64),
    ]
)
Acluster = Annotated[npt.NDArray[Any], acluster]
ocluster = np.dtype(
    [
        ("type", np.int64),
        ("x", np.float64),
        ("y", np.float64),
        ("z", np.float64),
        ("size", np.int64),
    ]
)
Ocluster = Annotated[npt.NDArray[Any], ocluster]
# region SRIM
trimdat = np.dtype(
    [
        ("name", str),
        ("atomic_number", int),
        ("energy", float),
        ("pos", float, 3),
        ("dir", float, 3),
    ]
)
Trimdat = Annotated[npt.NDArray[Any], trimdat]
